# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A widget to enter a keyboard shortcut.

Loosely based on kkeysequencewidget.cpp from KDE :-)

"""

from __future__ import unicode_literals

from PyQt4.QtCore import QEvent, QSize, Qt, QTimer, pyqtSignal
from PyQt4.QtGui import (
    QApplication, QHBoxLayout, QKeySequence, QPushButton, QToolButton, QWidget)

import app
import icons

from . import ClearButton


class KeySequenceWidget(QWidget):

    keySequenceChanged = pyqtSignal(int)
    
    def __init__(self, parent=None, num=0):
        super(KeySequenceWidget, self).__init__(parent)
        self._num = num
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        self.setLayout(layout)
        
        self.button = KeySequenceButton(self)
        self.clearButton = ClearButton(self, iconSize=QSize(16, 16))
        layout.addWidget(self.button)
        layout.addWidget(self.clearButton)
        
        self.clearButton.clicked.connect(self.clear)
        app.translateUI(self)
        
    def translateUI(self):
        self.button.setToolTip(_("Start recording a key sequence."))
        self.clearButton.setToolTip(_("Clear the key sequence."))
        
    def setShortcut(self, shortcut):
        """Sets the initial shortcut to display."""
        self.button.setKeySequence(shortcut)
    
    def shortcut(self):
        """Returns the currently set key sequence."""
        return self.button.keySequence()
        
    def clear(self):
        """Empties the displayed shortcut."""
        if self.button.isRecording():
            self.button.cancelRecording()
        if not self.button.keySequence().isEmpty():
            self.button.setKeySequence(QKeySequence())
            self.keySequenceChanged.emit(self._num)

    def setModifierlessAllowed(self, allow):
        self.button._modifierlessAllowed = allow
        
    def isModifierlessAllowed(self):
        return self.button._modifierlessAllowed
    
    def num(self):
        return self._num


class KeySequenceButton(QPushButton):
    
    def __init__(self, parent=None):
        super(KeySequenceButton, self).__init__(parent)
        self.setIcon(icons.get("configure"))
        self._modifierlessAllowed = False
        self._seq = QKeySequence()
        self._timer = QTimer()
        self._timer.setSingleShot(True)
        self._isrecording = False
        self.clicked.connect(self.startRecording)
        self._timer.timeout.connect(self.doneRecording)

    def setKeySequence(self, seq):
        self._seq = seq
        self.updateDisplay()

    def keySequence(self):
        if self._isrecording:
            self.doneRecording()
        return self._seq
    
    def updateDisplay(self):
        if self._isrecording:
            s = self._recseq.toString(QKeySequence.NativeText).replace('&', '&&')
            if self._modifiers:
                if s: s += ","
                s += QKeySequence(self._modifiers).toString(QKeySequence.NativeText)
            elif self._recseq.isEmpty():
                s = _("Input")
            s += " ..."
        else:
            s = self._seq.toString(QKeySequence.NativeText).replace('&', '&&')
        self.setText(s)

    def isRecording(self):
        return self._isrecording
        
    def event(self, ev):
        if self._isrecording:
            # prevent Qt from special casing Tab and Backtab
            if ev.type() == QEvent.KeyPress:
                self.keyPressEvent(ev)
                return True
        return super(KeySequenceButton, self).event(ev)
        
    def keyPressEvent(self, ev):
        if not self._isrecording:
            return super(KeySequenceButton, self).keyPressEvent(ev)
        if ev.isAutoRepeat():
            return
        modifiers = int(ev.modifiers() & (Qt.SHIFT | Qt.CTRL | Qt.ALT | Qt.META))
        ev.accept()
        
        key = ev.key()
        # check if key is a modifier or a character key without modifier (and if that is allowed)
        if (
            # don't append the key if the key is -1 (garbage) or a modifier ...
            key not in (-1, Qt.Key_AltGr, Qt.Key_Shift, Qt.Key_Control,
                            Qt.Key_Alt, Qt.Key_Meta, Qt.Key_Menu)
            # or if this is the first key and without modifier and modifierless keys are not allowed
            and (self._modifierlessAllowed
                 or self._recseq.count() > 0
                 or modifiers & ~Qt.SHIFT
                 or not ev.text()
                 or (modifiers & Qt.SHIFT
                     and key in (Qt.Key_Return, Qt.Key_Space, Qt.Key_Tab, Qt.Key_Backtab,
                                 Qt.Key_Backspace, Qt.Key_Delete, Qt.Key_Escape)))):
            # change Shift+Backtab into Shift+Tab
            if key == Qt.Key_Backtab and modifiers & Qt.SHIFT:
                key = Qt.Key_Tab | modifiers
            # remove the Shift modifier if it doesn't make sense
#            elif (Qt.Key_Exclam <= key <= Qt.Key_At
#                  or Qt.Key_Z < key <= 0x0ff):
#                key = key | (modifiers & ~Qt.SHIFT)
            else:
                key = key | modifiers
            
            # append max. 4 keystrokes
            if self._recseq.count() < 4:
                l = list(self._recseq)
                l.append(key)
                self._recseq = QKeySequence(*l)
        
        self._modifiers = modifiers
        self.controlTimer()
        self.updateDisplay()
        
    def keyReleaseEvent(self, ev):
        if not self._isrecording:
            return super(KeySequenceButton, self).keyReleaseEvent(ev)
        modifiers = int(ev.modifiers() & (Qt.SHIFT | Qt.CTRL | Qt.ALT | Qt.META))
        ev.accept()
        
        self._modifiers = modifiers
        self.controlTimer()
        self.updateDisplay()
    
    def hideEvent(self, ev):
        if self._isrecording:
            self.cancelRecording()
        super(KeySequenceButton, self).hideEvent(ev)
        
    def controlTimer(self):
        if self._modifiers or self._recseq.isEmpty():
            self._timer.stop()
        else:
            self._timer.start(600)
    
    def startRecording(self):
        self.setFocus(True) # because of QTBUG 17810
        self.setDown(True)
        self.setStyleSheet("text-align: left;")
        self._isrecording = True
        self._recseq = QKeySequence()
        self._modifiers = int(QApplication.keyboardModifiers() & (Qt.SHIFT | Qt.CTRL | Qt.ALT | Qt.META))
        self.grabKeyboard()
        self.updateDisplay()
        
    def doneRecording(self):
        self._seq = self._recseq
        self.cancelRecording()
        self.clearFocus()
        self.parentWidget().keySequenceChanged.emit(self.parentWidget().num())
        
    def cancelRecording(self):
        if not self._isrecording:
            return
        self.setDown(False)
        self.setStyleSheet(None)
        self._isrecording = False
        self.releaseKeyboard()
        self.updateDisplay()





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
This module monitors open documents for changes on disk (overwrite or move and
delete operations) by external programs.

The documentwatcher module is used to do the actual monitoring work,
this module checks if a touched file really changed and pops up the window
if needed.

"""

from __future__ import unicode_literals


from PyQt4.QtCore import QSettings, QTimer


def enabled():
    """Return whether watching documents is enabled by the user."""
    return QSettings().value("externalchanges/enabled", True, bool)


def setEnabled(enable):
    """Enable or disable watching documents, saves the setting."""
    if enabled() != bool(enable):
        QSettings().setValue("externalchanges/enabled", bool(enable))
        setup()


def setup():
    """Start or stop the document watching according to the settings."""
    import documentwatcher
    if enabled():
        documentwatcher.documentChangedOnDisk.connect(slotDocumentChanged)
        documentwatcher.start()
    else:
        documentwatcher.documentChangedOnDisk.disconnect(slotDocumentChanged)
        documentwatcher.stop()


def changedDocuments():
    """Return a list of really changed Documents.
    
    When a document is not modified and the file on disk is exactly the same,
    the document is not considered having been changed on disk.
    
    """
    import documentwatcher
    for w in documentwatcher.DocumentWatcher.instances():
        d = w.document()
        if w.changed and not d.isModified():
            filename = d.url().toLocalFile()
            if filename:
                try:
                    with open(filename, 'rb') as diskfile:
                        if diskfile.read() == d.encodedText():
                            w.changed = False
                except (OSError, IOError):
                    pass
    return [w.document() for w in documentwatcher.DocumentWatcher.instances()
              if w.changed]


def display(documents):
    """Diplay the window showing the specified Documents."""
    from . import widget
    window = widget.window()
    window.setDocuments(documents)
    window.show()


def displayChangedDocuments():
    """Display the window, even if there are no changed files."""
    display(changedDocuments())


def checkChangedDocuments():
    """Display the window if there are changed files."""
    docs = changedDocuments()
    if docs:
        display(docs)


# timer to wait before really looking at the changed files, a file could
# probably still be changing.
_timer = QTimer(singleShot=True, timeout=checkChangedDocuments)


def slotDocumentChanged(document):
    """Called when a document is changed."""
    _timer.start(500)


# initial setup
if enabled():
    setup()




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2011 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Analyze text to determine suitable completions.
"""

from __future__ import unicode_literals

import re
import os

import ly.lex as lx
import ly.lex.lilypond as lp
import ly.lex.scheme as scm
import ly.words
import tokeniter

from . import completiondata
from . import documentdata


class Analyzer(object):
    """Analyzes text at some cursor position and gives suitable completions."""
    def analyze(self, cursor):
        """Do the analyzing work; set the attributes column and model."""
        self.cursor = cursor
        block = cursor.block()
        self.column = column = cursor.position() - block.position()
        self.text = text = block.text()[:column]
        self.model = None
        
        # make a list of tokens exactly ending at the cursor position
        # and let state follow
        state = self.state = tokeniter.state(block)
        tokens = self.tokens = []
        for t in tokeniter.tokens(cursor.block()):
            if t.end > column:
                # cut off the last token and run the parser on it
                tokens.extend(state.tokens(text, t.pos))
                break
            tokens.append(t)
            state.follow(t)
            if t.end == column:
                break
        
        self.last = tokens[-1] if tokens else ''
        self.lastpos = self.last.pos if self.last else column
        
        parser = state.parser()
        
        # Map the parser class to a group of tests to return the model.
        # Try the tests until a model is returned.
        try:
            tests = self.tests[parser.__class__]
        except KeyError:
            return
        else:
            for function in tests:
                model = function(self)
                if model:
                    self.model = model
                    return

    def completions(self, cursor):
        """Analyzes text at cursor and returns a tuple (position, model).
        
        The position is an integer specifying the column in the line where the last
        text starts that should be completed.
        
        The model list the possible completions. If the model is None, there are no
        suitable completions.
        
        This function does its best to return extremely meaningful completions
        for the context the cursor is in.
        
        """
        self.analyze(cursor)
        return self.column, self.model
    
    def document_cursor(self):
        """Return the current QTextCursor, to harvest info from its document.
        
        By default this is simply the cursor given on analyze() or completions()
        but you can override this method to provide another cursor. This can
        be useful when the completion occurs in a small QTextDocument, which is
        in fact a part of the main document.
        
        """
        return self.cursor
    
    def tokenclasses(self):
        """Return the list of classes of the tokens."""
        return list(map(type, self.tokens))

    def backuntil(self, *classes):
        """Move self.column back until a token of *classes is encountered."""
        for t in self.tokens[::-1]:
            if isinstance(t, classes):
                break
            self.column = t.pos

    # Test functions that return a model or None
    def toplevel(self):
        """LilyPond toplevel document contents."""
        self.backuntil(lx.Space)
        return completiondata.lilypond_toplevel
        # maybe: check if behind \version or \language

    def book(self):
        """\\book {"""
        self.backuntil(lx.Space)
        cursor = self.document_cursor()
        return documentdata.doc(cursor.document()).bookcommands(cursor)
        
    def bookpart(self):
        """\\bookpart {"""
        self.backuntil(lx.Space)
        cursor = self.document_cursor()
        return documentdata.doc(cursor.document()).bookpartcommands(cursor)

    def score(self):
        """\\score {"""
        self.backuntil(lx.Space)
        cursor = self.document_cursor()
        return documentdata.doc(cursor.document()).scorecommands(cursor)

    def tweak(self):
        """complete property after \\tweak"""
        try:
            i = self.tokens.index('\\tweak')
        except ValueError:
            return
        tokens = self.tokens[i+1:]
        tokenclasses = self.tokenclasses()[i+1:]
        if tokenclasses == [lx.Space, lp.SchemeStart]:
            self.column -= 1
            return completiondata.lilypond_all_grob_properties
        elif tokenclasses == [lx.Space, lp.SchemeStart, scm.Quote]:
            self.column -= 2
            return completiondata.lilypond_all_grob_properties
        elif tokenclasses[:-1] == [lx.Space, lp.SchemeStart, scm.Quote]:
            self.column = self.lastpos - 2
            return completiondata.lilypond_all_grob_properties
        # 2.18-style [GrobName.]propertyname tweak
        if lp.GrobName in tokenclasses:
            self.backuntil(lx.Space, lp.DotPath)
            return completiondata.lilypond_grob_properties(tokens[1], False)
        if tokens:
            self.backuntil(lx.Space)
            return completiondata.lilypond_all_grob_properties_and_grob_names

    def key(self):
        """complete mode argument of '\\key'"""
        tokenclasses = self.tokenclasses()
        if '\\key' in self.tokens[-5:-2] and lp.Note in tokenclasses[-3:]:
            if self.last.startswith('\\'):
                self.column = self.lastpos
            return completiondata.lilypond_modes

    def clef(self):
        """complete \\clef names"""
        if '\\clef' in self.tokens[-4:-1]:
            self.backuntil(lx.Space, lp.StringQuotedStart)
            return completiondata.lilypond_clefs
            
    def repeat(self):
        """complete \\repeat types"""
        if '\\repeat' in self.tokens[-4:-1]:
            self.backuntil(lx.Space, lp.StringQuotedStart)
            return completiondata.lilypond_repeat_types

    def language(self):
        """complete \\language "name" """
        if '\\language' in self.tokens[-4:-1]:
            self.backuntil(lp.StringQuotedStart)
            return completiondata.language_names

    def include(self):
        """complete \\include """
        if '\\include' in self.tokens[-4:-2]:
            self.backuntil(lp.StringQuotedStart)
            dir = self.last[:self.last.rfind(os.sep)] if os.sep in self.last else None
            cursor = self.document_cursor()
            return documentdata.doc(cursor.document()).includenames(cursor, dir)

    def general_music(self):
        """fall back: generic music commands and user-defined commands."""
        if self.last.startswith('\\'):
            self.column = self.lastpos
        cursor = self.document_cursor()
        return documentdata.doc(cursor.document()).musiccommands(cursor)

    def lyricmode(self):
        """Commands inside lyric mode."""
        if self.last.startswith('\\'):
            self.column = self.lastpos
        cursor = self.document_cursor()
        return documentdata.doc(cursor.document()).lyriccommands(cursor)
        
    def music_glyph(self):
        r"""Complete \markup \musicglyph names."""
        try:
            i = self.tokens.index('\\musicglyph', -5, -3)
        except ValueError:
            return
        for t, cls in zip(self.tokens[i:], (
            lp.MarkupCommand, lx.Space, lp.SchemeStart, scm.StringQuotedStart, scm.String)):
            if type(t) is not cls:
                return
        if i + 4 < len(self.tokens):
            self.column = self.tokens[i + 4].pos
        return completiondata.music_glyphs

    def midi_instrument(self):
        """Complete midiInstrument = #"... """
        try:
            i = self.tokens.index('midiInstrument', -7, -2)
        except ValueError:
            return
        if self.last != '"':
            self.column = self.lastpos
        return completiondata.midi_instruments
        
    def font_name(self):
        """Complete #'font-name = #"..."""
        try:
            i = self.tokens.index('font-name', -7, -3)
        except ValueError:
            return
        if self.last != '"':
            self.column = self.lastpos
        return completiondata.font_names()
        
    def scheme_word(self):
        """Complete scheme word from scheme functions, etc."""
        if isinstance(self.last, scm.Word):
            self.column = self.lastpos
            cursor = self.document_cursor()
            return documentdata.doc(cursor.document()).schemewords()
        
    def markup(self):
        """\\markup {"""
        if self.last.startswith('\\'):
            if (self.last[1:] not in ly.words.markupcommands
                and self.last != '\\markup'):
                self.column = self.lastpos
            else:
                return completiondata.lilypond_markup_commands
        else:
            m = re.search(r'\w+$', self.last)
            if m:
                self.column = self.lastpos + m.start()
        cursor = self.document_cursor()
        return documentdata.doc(cursor.document()).markup(cursor)
        
    def header(self):
        """\\header {"""
        if '=' in self.tokens[-3:] or self.last.startswith('\\'):
            if self.last.startswith('\\'):
                self.column = self.lastpos
            return completiondata.lilypond_markup
        if self.last[:1].isalpha():
            self.column = self.lastpos
        return completiondata.lilypond_header_variables

    def paper(self):
        """\\paper {"""
        if '=' in self.tokens[-3:] or self.last.startswith('\\'):
            if self.last.startswith('\\'):
                self.column = self.lastpos
            return completiondata.lilypond_markup
        if self.last[:1].isalpha():
            self.column = self.lastpos
        return completiondata.lilypond_paper_variables
        
    def layout(self):
        """\\layout {"""
        self.backuntil(lx.Space)
        return completiondata.lilypond_layout_variables

    def midi(self):
        """\\midi {"""
        self.backuntil(lx.Space)
        return completiondata.lilypond_midi_variables

    def engraver(self):
        """Complete engraver names."""
        cmd_in = lambda tokens: '\\remove' in tokens or '\\consists' in tokens
        if isinstance(self.state.parser(), lp.ParseString):
            if not cmd_in(self.tokens[-5:-2]):
                return
            if self.last != '"':
                if '"' not in self.tokens[-2:-1]:
                    return
                self.column = self.lastpos
            return completiondata.lilypond_engravers
        if cmd_in(self.tokens[-3:-1]):
            self.backuntil(lx.Space)
            return completiondata.lilypond_engravers
        
    def context_variable_set(self):
        if '=' in self.tokens[-4:]:
            if isinstance(self.last, scm.Word):
                self.column = self.lastpos
                cursor = self.document_cursor()
                return documentdata.doc(cursor.document()).schemewords()
            if self.last.startswith('\\'):
                self.column = self.lastpos
            return completiondata.lilypond_markup

    def context(self):
        self.backuntil(lx.Space)
        return completiondata.lilypond_context_contents

    def with_(self):
        self.backuntil(lx.Space)
        return completiondata.lilypond_with_contents

    def translator(self):
        """complete context name after \\new, \\change or \\context in music"""
        for t in self.tokens[-2::-1]:
            if isinstance(t, lp.ContextName):
                return
            elif isinstance(t, lp.Translator):
                break
        self.backuntil(lx.Space)
        return completiondata.lilypond_contexts

    def override(self):
        """\\override and \\revert"""
        tokenclasses = self.tokenclasses()
        try:
            # check if there is a GrobName in the last 5 tokens
            i = tokenclasses.index(lp.GrobName, -5)
        except ValueError:
            # not found, then complete Contexts and or Grobs
            # (only if we are in the override parser and there's no "=")
            if isinstance(self.state.parser(), scm.ParseScheme):
                return
            self.backuntil(lp.DotPath, lx.Space)
            if (isinstance(self.state.parsers()[1], (
                    lp.ParseWith,
                    lp.ParseContext,
                    ))
                or lp.DotPath in tokenclasses):
                return completiondata.lilypond_grobs
            return completiondata.lilypond_contexts_and_grobs
        # yes, there is a GrobName at i
        count = len(self.tokens) - i - 1 # tokens after grobname
        if count == 0:
            self.column = self.lastpos
            return completiondata.lilypond_grobs
        elif count >= 2:
            # set the place of the scheme-start "#" as the column
            self.column = self.tokens[i+2].pos
        test = [lx.Space, lp.SchemeStart, scm.Quote, scm.Word]
        if tokenclasses[i+1:] == test[:count]:
            return completiondata.lilypond_grob_properties(self.tokens[i])
        self.backuntil(lp.DotPath, lx.Space)
        return completiondata.lilypond_grob_properties(self.tokens[i], False)
    
    def revert(self):
        """test for \\revert in general music expressions
        
        (because the revert parser drops out of invalid constructs, which happen
        during typing).
        
        """
        if '\\revert' in self.tokens:
            return self.override()
    
    def set_unset(self):
        """\\set and \\unset"""
        tokenclasses = self.tokenclasses()
        self.backuntil(lx.Space, lp.DotPath)
        if lp.ContextProperty in tokenclasses and isinstance(self.last, lx.Space):
            return # fall back to music?
        elif lp.DotPath in tokenclasses:
            return completiondata.lilypond_context_properties
        return completiondata.lilypond_contexts_and_properties

    def markup_override(self):
        """test for \\markup \\override inside scheme"""
        try:
            i = self.tokens.index('\\override', -6, -4)
        except ValueError:
            return
        for t, cls in zip(self.tokens[i:], (
            lp.MarkupCommand, lx.Space, lp.SchemeStart, scm.Quote, scm.OpenParen)):
            if type(t) is not cls:
                return
        if len(self.tokens) > i + 5:
            self.column = self.lastpos
        return completiondata.lilypond_markup_properties

    def scheme_other(self):
        """test for other scheme words"""
        if isinstance(self.last, (
            lp.SchemeStart,
            scm.OpenParen,
            scm.Word,
            )):
            if isinstance(self.last, scm.Word):
                self.column = self.lastpos
            cursor = self.document_cursor()
            return documentdata.doc(cursor.document()).schemewords()
    
    def accidental_style(self):
        """test for \accidentalStyle"""
        try:
            i = self.tokens.index("\\accidentalStyle")
        except ValueError:
            return
        self.backuntil(lx.Space, lp.DotPath)
        tokens = self.tokens[i+1:]
        tokenclasses = self.tokenclasses()[i+1:]
        try:
            i = tokenclasses.index(lp.AccidentalStyleSpecifier)
        except ValueError:
            pass
        else:
            if lx.Space in tokenclasses[i+1:]:
                return
        if lp.ContextName in tokenclasses:
            return completiondata.lilypond_accidental_styles
        return completiondata.lilypond_accidental_styles_contexts

    def hide_omit(self):
        r"""test for \omit and \hide"""
        indices = []
        for t in "\\omit", "\\hide":
            try:
                indices.append(self.tokens.index(t, -6))
            except ValueError:
                pass
        if not indices:
            return
        self.backuntil(lx.Space, lp.DotPath)
        i = max(indices)
        tokens = self.tokens[i+1:]
        tokenclasses = self.tokenclasses()[i+1:]
        if lp.GrobName not in tokenclasses[:-1]:
            if lp.ContextName in tokenclasses:
                return completiondata.lilypond_grobs
            return completiondata.lilypond_contexts_and_grobs


    # Mapping from Parsers to the lists of functions to run.
    tests = {
        lp.ParseGlobal: (
            repeat,
            toplevel,
        ),
        lp.ParseBook: (
            book,
        ),
        lp.ParseBookPart: (
            bookpart,
        ),
        lp.ParseScore: (
            score,
        ),
        lp.ParseMusic: (
            tweak,
            scheme_word,
            key,
            clef,
            repeat,
            accidental_style,
            hide_omit,
            revert,
            general_music,
        ),
        lp.ParseNoteMode: (
            tweak,
            scheme_word,
            key,
            clef,
            repeat,
            accidental_style,
            hide_omit,
            revert,
            general_music,
        ),
        lp.ParseMarkup: (
            markup,
        ),
        lp.ParseHeader: (
            header,
        ),
        lp.ParsePaper: (
            paper,
        ),
        lp.ParseLayout: (
            accidental_style,
            hide_omit,
            layout,
        ),
        lp.ParseMidi: (
            midi,
        ),
        lp.ParseContext: (
            engraver,
            context_variable_set,
            context,
        ),
        lp.ParseWith: (
            engraver,
            context_variable_set,
            with_,
        ),
        lp.ParseTranslator: (
            translator,
        ),
        lp.ExpectTranslatorId: (
            translator,
        ),
        lp.ParseOverride: (
            override,
        ),
        lp.ParseRevert: (
            override,
        ),
        lp.ParseSet: (
            set_unset,
        ),
        lp.ParseUnset: (
            set_unset,
        ),
        lp.ParseTweak: (
            tweak,
        ),
        lp.ParseTweakGrobProperty: (
            tweak,
        ),
        lp.ParseString: (
            engraver,
            clef,
            repeat,
            midi_instrument,
            include,
            language,
        ),
        lp.ParseClef: (
            clef,
        ),
        lp.ParseRepeat: (
            repeat,
        ),
        scm.ParseScheme: (
            override,
            tweak,
            markup_override,
            scheme_other,
        ),
        scm.ParseString: (
            music_glyph,
            midi_instrument,
            font_name,
        ),
        lp.ParseLyricMode: (
            repeat,
            lyricmode,
        ),
        lp.ParseAccidentalStyle: (
            accidental_style,
        ),
        lp.ParseScriptAbbreviationOrFingering: (
            accidental_style,
        ),
        lp.ParseHideOmit: (
            hide_omit,
        ),
        lp.ParseGrobPropertyPath: (
            revert,
        ),
    }




# vimode -- Vi Mode package for QPlainTextEdit
#
# Copyright (c) 2012 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Normal ViMode.
"""

from __future__ import unicode_literals

import re

from PyQt4.QtCore import Qt
from PyQt4.QtGui import QTextCursor

from . import handlerbase

commands = []


class NormalMode(handlerbase.Handler):
    def __init__(self, vimode):
        super(NormalMode, self).__init__(vimode)
        self._command = []
        self._count = 0
        
    def enter(self):
        self.vimode.textEdit().setCursorWidth(0)
    
    def updateCursorPosition(self):
        cursor = self.textCursor()
        if cursor.hasSelection() and cursor.position() > cursor.anchor():
            cursor.clearSelection()
            cursor.movePosition(QTextCursor.PreviousCharacter, QTextCursor.KeepAnchor)
        else:
            cursor.clearSelection()
            cursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor)
        self.vimode.drawCursor(cursor)

    def handleKeyPress(self, ev):
        # is a number being entered?
        if not self._command and Qt.Key_0 <= ev.key() <= Qt.Key_9:
            self._count *= 10
            self._count += ev.key() - Qt.Key_0
            return True
        # backspace?
        if ev.key() == Qt.Key_Backspace:
            if self._command:
                # remove last key typed
                del self._command[-1]
            else:
                self._count = 0
            return True
        
        # TEMP
        if ev.text():
            self._command.append(ev.text())
        
        cmd = ''.join(self._command)
        for s, f in commands:
            if re.compile(s).match(cmd):
                f(self)
                self._count = 0
                self._command = []
                return True
        
        return True
    
    
    def command(cmds, count=True, set_cursor=True):
        def decorator(func):
            def wrapper(self):
                cursor = self.textCursor()
                if count:
                    for i in range(max(1, self._count)):
                        func(self, cursor)
                else:
                    func(self, cursor)
                if set_cursor:
                    self.setTextCursor(cursor)
            for c in cmds:
                commands.append((c, wrapper))
        return decorator
    
    @command(['h', '<left>'])
    def move_left(self, cursor):
        if cursor.position() > cursor.block().position():
            cursor.movePosition(QTextCursor.PreviousCharacter)
    
    @command(['l', '<right>'])
    def move_right(self, cursor):
        if cursor.position() < cursor.block().position() + cursor.block().length() - 2:
            cursor.movePosition(QTextCursor.NextCharacter)
    
    @command('i', set_cursor=False)
    def insert_mode(self, cursor):
        self.setInsertMode()
    
    
        




# This module provides the same api via ctypes as John Harrison's pyrex-based
# PortMIDI binding. 
# Don't use this module directly but via the toplevel API of this package.
# Copyright (c) 2011 - 2014 by Wilbert Berendsen, placed in the public domain.

import array
from ctypes import byref, create_string_buffer

from .pm_ctypes import (
    libpm, libpt,
    pmHostError, PmEvent,
    PmTimeProcPtr, NullTimeProcPtr,
    PortMidiStreamPtr,
)

from . import MidiException


__all__ = [
    'TRUE',
    'FALSE',
    'Initialize',
    'Terminate',
    'CountDevices',
    'GetDeviceInfo',
    'GetDefaultInputDeviceID',
    'GetDefaultOutputDeviceID',
    'GetErrorText',
    'Time',
    'Input',
    'Output',
]


FALSE = 0
TRUE = 1


def Initialize():
    libpm.Pm_Initialize()
    # equiv to TIME_START: start timer w/ ms accuracy
    libpt.Pt_Start(1, NullTimeProcPtr, None)


def Terminate():
    libpm.Pm_Terminate()


def GetDeviceInfo(device_id):
    info_ptr = libpm.Pm_GetDeviceInfo(device_id)
    if info_ptr:
        info = info_ptr.contents
        return (
            info.interf,
            info.name,
            bool(info.input),
            bool(info.output),
            bool(info.opened),
        )


CountDevices = libpm.Pm_CountDevices
GetDefaultInputDeviceID = libpm.Pm_GetDefaultInputDeviceID
GetDefaultOutputDeviceID = libpm.Pm_GetDefaultOutputDeviceID
GetErrorText = libpm.Pm_GetErrorText
Time = libpt.Pt_Time


class Output(object):
    buffer_size = 1024
    def __init__(self, device_id, latency=0):
        self.device_id = device_id
        self.latency = latency
        self._midi_stream = PortMidiStreamPtr()
        self._open = False
        self._open_device()

    def _open_device(self):
        err = libpm.Pm_OpenOutput(byref(self._midi_stream), self.device_id,
            None, 0, NullTimeProcPtr, None, self.latency)
        _check_error(err)
        self._open = True

    def Close(self):
        if self._open and GetDeviceInfo(self.device_id)[4]:
            err = libpm.Pm_Abort(self._midi_stream)
            _check_error(err)
            err = libpm.Pm_Close(self._midi_stream)
            _check_error(err)
            self._open = False
    
    __del__ = Close

    def Write(self, data):
        bufsize = self.buffer_size

        if len(data) > bufsize:
            raise ValueError("too much data for buffer")

        BufferType = PmEvent * bufsize
        buf = BufferType()

        for i, event in enumerate(data):
            msg, buf[i].timestamp = event
            if len(msg) > 4 or len(msg) < 1:
                raise ValueError("invalid message size")
            message = 0
            for j, byte in enumerate(msg):
                message += ((byte & 0xFF) << (8*j))
            buf[i].message = message
        err = libpm.Pm_Write(self._midi_stream, buf, len(data))
        _check_error(err)

    def WriteShort(self, status, data1=0, data2=0):
        buf = PmEvent()
        buf.timestamp = libpt.Pt_Time()
        buf.message = (((data2 << 16) & 0xFF0000) |
            ((data1 << 8) & 0xFF00) | (status & 0xFF))
        err = libpm.Pm_Write(self._midi_stream, buf, 1)
        _check_error(err)

    def WriteSysEx(self, timestamp, msg):
        """msg may be a tuple or list of ints, or a bytes string."""
        if isinstance(msg, (tuple, list)):
            msg = array.array('B', msg).tostring()
        cur_time = Time()
        err = libpm.Pm_WriteSysEx(self._midi_stream, timestamp, msg)
        _check_error(err)
        while Time() == cur_time:
            pass


class Input(object):
    def __init__(self, device_id, bufsize=1024):
        self.device_id = device_id
        self.buffer_size = bufsize
        self._midi_stream = PortMidiStreamPtr()
        self._open = False
        self._open_device()

    def _open_device(self):
        err = libpm.Pm_OpenInput(byref(self._midi_stream), self.device_id,
            None, 100, NullTimeProcPtr, None)
        _check_error(err)
        self._open = True

    def Close(self):
        """Closes a midi stream, flushing any pending buffers."""
        if self._open and GetDeviceInfo(self.device_id)[4]:
            err = libpm.Pm_Close(self._midi_stream)
            _check_error(err)
            self._open = False
        
    __del__ = Close

    def Poll(self):
        return libpm.Pm_Poll(self._midi_stream)

    def Read(self, length):
        bufsize = self.buffer_size
        BufferType = PmEvent * bufsize
        buf = BufferType()

        if not 1 <= length <= bufsize:
            raise ValueError("invalid length")
        num_events = libpm.Pm_Read(self._midi_stream, buf, length)
        _check_error(num_events)

        data = []
        for i in range(num_events):
            ev = buf[i]
            msg = ev.message
            msg = (msg & 255, (msg>>8) & 255, (msg>>16) & 255, (msg>>24) & 255)
            data.append((msg, ev.timestamp))
        return data


def _check_error(err_no):
    if err_no < 0:
        if err_no == pmHostError:
            err_msg = create_string_buffer(b'\000' * 256)
            libpm.Pm_GetHostErrorText(err_msg, 256)
            err_msg = err_msg.value
        else:
            err_msg = libpm.Pm_GetErrorText(err_no)
        raise MidiException(
            "PortMIDI-ctypes error [{0}]: {1}".format(err_no,
                                                      err_msg.decode('utf-8')))





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A JobManager exists for every Document, and ensures no two jobs are running
at the same time.

It also sends the app-wide signals jobStarted() and jobFinished().

"""

from __future__ import unicode_literals

import app
import plugin
import signals


def manager(document):
    return JobManager.instance(document)


def job(document):
    return JobManager.instance(document).job()


def is_running(document):
    if job(document):
        return job(document).is_running()
    return False


class JobManager(plugin.DocumentPlugin):
    
    started = signals.Signal()  # Job
    finished = signals.Signal() # Job, success
    
    def __init__(self, document):
        self._job = None
        
    def start_job(self, job):
        """Starts a Job on our behalf."""
        if not self.is_running():
            self._job = job
            job.done.connect(self._finished)
            job.start()
            self.started(job)
            app.jobStarted(self.document(), job)
        
    def _finished(self, success):
        self.finished(job, success)
        app.jobFinished(self.document(), self._job, success)
    
    def job(self):
        """Returns the last job if any."""
        return self._job

    def is_running(self):
        """Returns True when a job is running."""
        if self._job:
            return self._job.is_running() and not self._job.is_aborted()






"""
MIDI input controls
"""

from __future__ import unicode_literals

import weakref

from PyQt4.QtCore import QSettings, Qt
from PyQt4.QtGui import (QCheckBox, QComboBox, QGridLayout, QGroupBox,
                         QHBoxLayout, QLabel, QRadioButton, QToolButton,
                         QVBoxLayout, QWidget)

import app

import midiinput


class Widget(QWidget):
    def __init__(self, dockwidget):
        super(Widget, self).__init__(dockwidget)
        self._document = None
        self._midiin = midiinput.MidiIn(self)
        self._dockwidget = weakref.ref(dockwidget)
        
        signals = list()
        
        self._labelmidichannel = QLabel()
        self._midichannel = QComboBox()
        signals.append(self._midichannel.currentIndexChanged)
        
        self._labelkeysignature = QLabel()
        self._keysignature = QComboBox()
        signals.append(self._keysignature.currentIndexChanged)
        
        self._labelaccidentals = QLabel()
        self._accidentalssharps = QRadioButton()
        signals.append(self._accidentalssharps.clicked)
        self._accidentalsflats = QRadioButton()
        signals.append(self._accidentalsflats.clicked)
        self._groupaccidentals = QGroupBox()
        self._groupaccidentals.setFlat(True)
        hbox = QHBoxLayout()
        self._groupaccidentals.setLayout(hbox)
        hbox.addWidget(self._accidentalssharps)
        hbox.addWidget(self._accidentalsflats)
        self._accidentalssharps.setChecked(True)
        
        self._chordmode = QCheckBox()
        signals.append(self._chordmode.clicked)
        
        self._relativemode = QCheckBox()
        signals.append(self._relativemode.clicked)

        self._labeldamper = QLabel()
        self._damper = QComboBox()
        
        self._labelsostenuto = QLabel()
        self._sostenuto = QComboBox()
        
        self._labelsoft = QLabel()
        self._soft = QComboBox()
        
        ac = self.parentWidget().actionCollection
        self._capture = QToolButton()
        self._capture.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self._capture.setDefaultAction(ac.capture_start)
        
        self._notemode = QLabel()
        
        layout = QVBoxLayout()
        self.setLayout(layout)
        grid = QGridLayout(spacing=0)
        layout.addLayout(grid)
               
        grid.addWidget(self._labelmidichannel, 0, 0)
        grid.addWidget(self._midichannel, 0, 1)
        grid.addWidget(self._labelkeysignature, 1, 0)
        grid.addWidget(self._keysignature, 1, 1)
        grid.addWidget(self._labelaccidentals, 2, 0)
        grid.addWidget(self._groupaccidentals, 2, 1)
        grid.addWidget(self._chordmode, 3, 0)
        grid.addWidget(self._relativemode, 3, 1)
        grid.addWidget(self._labeldamper, 4, 0)
        grid.addWidget(self._damper, 4, 1)
        grid.addWidget(self._labelsostenuto, 5, 0)
        grid.addWidget(self._sostenuto, 5, 1)
        grid.addWidget(self._labelsoft, 6, 0)
        grid.addWidget(self._soft, 6, 1)
        
        hbox = QHBoxLayout()
        layout.addLayout(hbox)
        hbox.addWidget(self._capture)
        hbox.addStretch()
        
        app.translateUI(self)
        
        self.loadsettings()
        for s in signals:
            s.connect(self.savesettings)
    
    def mainwindow(self):
        return self._dockwidget().mainwindow()
    
    def channel(self):
        return self._midichannel.currentIndex()
    
    def keysignature(self):
        return self._keysignature.currentIndex()
    
    def accidentals(self):
        if self._accidentalsflats.isChecked():
            return 'flats'
        else:
            return 'sharps'
    
    def chordmode(self):
        return self._chordmode.isChecked()
    
    def relativemode(self):
        return self._relativemode.isChecked()

    def startcapturing(self):
        self._midiin.capture()
        ac = self.parentWidget().actionCollection
        while self._capture.actions():    # remove all old actions
            self._capture.removeAction(self._capture.actions()[0])
        self._capture.setDefaultAction(ac.capture_stop)
    
    def stopcapturing(self):
        self._midiin.capturestop()
        ac = self.parentWidget().actionCollection
        while self._capture.actions():    # remove all old actions
            self._capture.removeAction(self._capture.actions()[0])
        self._capture.setDefaultAction(ac.capture_start)
    
    def savesettings(self):
        s = QSettings()
        s.beginGroup("midiinputdock")
        s.setValue("midichannel", self._midichannel.currentIndex())
        s.setValue("keysignature", self._keysignature.currentIndex())
        if self._accidentalsflats.isChecked():
            s.setValue("accidentals", 'flats')
        else:
            s.setValue("accidentals", 'sharps')
        s.setValue("chordmode", self._chordmode.isChecked())
        s.setValue("relativemode", self._relativemode.isChecked())
    
    def loadsettings(self):
        s = QSettings()
        s.beginGroup("midiinputdock")
        self._midichannel.setCurrentIndex(s.value("midichannel", 0, int))
        self._keysignature.setCurrentIndex(s.value("keysignature", 7, int))
        if s.value("accidentals", 'sharps', str) == 'flats':
            self._accidentalsflats.setChecked(True)
        else:
            self._accidentalssharps.setChecked(True)
        self._chordmode.setChecked(s.value("chordmode", False, bool))
        self._relativemode.setChecked(s.value("relativemode", False, bool))

    def translateUI(self):
        self._labelmidichannel.setText(_("MIDI channel"))
        self._midichannel.addItems([_("all")]+[str(i) for i in range(1,17)])
        self._labelkeysignature.setText(_("Key signature"))
        self._keysignature.addItems([
            _("C flat major (7 flats)"),
            _("G flat major (6 flats)"),
            _("D flat major (5 flats)"),
            _("A flat major (4 flats)"),
            _("E flat major (3 flats)"),
            _("B flat major (2 flats)"),
            _("F major (1 flat)"),
            _("C major"),
            _("G major (1 sharp)"),
            _("D major (2 sharps)"),
            _("A major (3 sharps)"),
            _("E major (4 sharps)"),
            _("B major (5 sharps)"),
            _("F sharp major (6 sharps)"),
            _("C sharp major (7 sharps)")
            ])
        self._keysignature.setCurrentIndex(7)
        self._labelaccidentals.setText(_("Accidentals"))
        self._accidentalssharps.setText(_("sharps"))
        self._accidentalsflats.setText(_("flats"))
        self._chordmode.setText(_("Chord mode"))
        self._chordmode.setToolTip(_(
            "Enter simultaneously played notes as chords. "
            "See \"What's This\" for more information."))
        self._chordmode.setWhatsThis(_(
            "Notes which are played simultaneously are written "
            "as chords. As a consequence they are not written "
            "before the last key is lifted. Of course single "
            "can also be entered."))
        self._relativemode.setText(_("Relative mode"))
        self._relativemode.setToolTip(_(
            "Enter octaves of notes relative to the last note. "
            "See \"What's This\" for more information."))
        self._relativemode.setWhatsThis(_(
            "Enter octaves of notes relative to the last note. "
            "This refers to the last key pressed on the MIDI keyboard, not the last note in the document."
            "Hold Shift with a note to enter an octave check."))
        self._labeldamper.setText(_("Damper pedal"))
        self._labelsostenuto.setText(_("Sostenuto pedal"))
        self._labelsoft.setText(_("Soft pedal"))



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The log dockwindow.
"""

from __future__ import unicode_literals

from PyQt4.QtCore import QSettings, Qt
from PyQt4.QtGui import QAction, QKeySequence

import actioncollection
import actioncollectionmanager
import app
import panel


class LogTool(panel.Panel):
    """A dockwidget showing the log of running Jobs."""
    def __init__(self, mainwindow):
        super(LogTool, self).__init__(mainwindow)
        self.hide()
        self.toggleViewAction().setShortcut(QKeySequence("Meta+Alt+L"))
        ac = self.actionCollection = Actions()
        ac.log_next_error.triggered.connect(self.slotNextError)
        ac.log_previous_error.triggered.connect(self.slotPreviousError)
        actioncollectionmanager.manager(mainwindow).addActionCollection(ac)
        mainwindow.addDockWidget(Qt.BottomDockWidgetArea, self)
        app.jobStarted.connect(self.slotJobStarted)
        app.jobFinished.connect(self.slotJobFinished)
    
    def translateUI(self):
        self.setWindowTitle(_("LilyPond Log"))
        self.toggleViewAction().setText(_("LilyPond &Log"))
        
    def createWidget(self):
        from . import logwidget
        return logwidget.LogWidget(self)
    
    def slotJobStarted(self, doc, job):
        """Called whenever job starts, decides whether to follow it and show the log."""
        import jobattributes
        jattrs = jobattributes.get(job)
        if doc == self.mainwindow().currentDocument() or self.mainwindow() == jattrs.mainwindow:
            self.widget().switchDocument(doc)
            if not jattrs.hidden and QSettings().value("log/show_on_start", True, bool):
                self.show()

    def slotJobFinished(self, document, job, success):
        import jobattributes
        if (not success and not job.is_aborted()
                and not jobattributes.get(job).hidden
                and document == self.mainwindow().currentDocument()):
            self.show()
    
    def slotNextError(self):
        """Jumps to the position pointed to by the next error message."""
        self.activate()
        self.widget().gotoError(1)
    
    def slotPreviousError(self):
        """Jumps to the position pointed to by the next error message."""
        self.activate()
        self.widget().gotoError(-1)
        

class Actions(actioncollection.ActionCollection):
    name = "logtool"
    def createActions(self, parent=None):
        self.log_next_error = QAction(parent)
        self.log_previous_error = QAction(parent)
        
        self.log_next_error.setShortcut(QKeySequence("Ctrl+E"))
        self.log_previous_error.setShortcut(QKeySequence("Ctrl+Shift+E"))
        
    def translateUI(self):
        self.log_next_error.setText(_("Next Error Message"))
        self.log_previous_error.setText(_("Previous Error Message"))


# log errors by initializing Errors instance
@app.jobStarted.connect
def _log_errors(document):
    from . import errors
    errors.errors(document)




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The Parts widget.
"""

from __future__ import unicode_literals

from PyQt4.QtCore import Qt
from PyQt4.QtGui import (QGroupBox, QHBoxLayout, QLabel, QPushButton, QSplitter,
                         QStackedWidget, QToolButton, QTreeView, QVBoxLayout,
                         QWidget)

import app
import icons
import widgets.treewidget

from . import parts


class ScorePartsWidget(QSplitter):
    def __init__(self, parent):
        super(ScorePartsWidget, self).__init__(parent)
        
        self.typesLabel = QLabel()
        self.typesView = QTreeView(
            selectionMode=QTreeView.ExtendedSelection,
            selectionBehavior=QTreeView.SelectRows,
            animated=True,
            headerHidden=True)
        self.scoreLabel = QLabel()
        self.scoreView = widgets.treewidget.TreeWidget(
            selectionMode=QTreeView.ExtendedSelection,
            selectionBehavior=QTreeView.SelectRows,
            headerHidden=True,
            animated=True,
            dragDropMode=QTreeView.InternalMove)
        self.addButton = QPushButton(icon = icons.get("list-add"))
        self.removeButton = QPushButton(icon = icons.get("list-remove"))
        self.upButton = QToolButton(icon = icons.get("go-up"))
        self.downButton = QToolButton(icon = icons.get("go-down"))
        self.partSettings = QStackedWidget()
        
        w = QWidget()
        self.addWidget(w)
        layout = QVBoxLayout(spacing=0)
        w.setLayout(layout)
        
        layout.addWidget(self.typesLabel)
        layout.addWidget(self.typesView)
        layout.addWidget(self.addButton)
        
        w = QWidget()
        self.addWidget(w)
        layout = QVBoxLayout(spacing=0)
        w.setLayout(layout)
        
        layout.addWidget(self.scoreLabel)
        layout.addWidget(self.scoreView)
        
        box = QHBoxLayout(spacing=0)
        layout.addLayout(box)
        
        box.addWidget(self.removeButton)
        box.addWidget(self.upButton)
        box.addWidget(self.downButton)
        
        self.addWidget(self.partSettings)

        self.typesView.setModel(parts.model())
        app.translateUI(self)
        
        # signal connections
        self.addButton.clicked.connect(self.slotAddButtonClicked)
        self.removeButton.clicked.connect(self.slotRemoveButtonClicked)
        self.typesView.doubleClicked.connect(self.slotDoubleClicked)
        self.scoreView.currentItemChanged.connect(self.slotCurrentItemChanged)
        self.upButton.clicked.connect(self.scoreView.moveSelectedChildrenUp)
        self.downButton.clicked.connect(self.scoreView.moveSelectedChildrenDown)
        
    def translateUI(self):
        bold = "<b>{0}</b>".format
        self.typesLabel.setText(bold(_("Available parts:")))
        self.scoreLabel.setText(bold(_("Score:")))
        self.addButton.setText(_("&Add"))
        self.removeButton.setText(_("&Remove"))
        self.upButton.setToolTip(_("Move up"))
        self.downButton.setToolTip(_("Move down"))

    def slotDoubleClicked(self, index):
        self.addParts([index])
        
    def slotAddButtonClicked(self):
        self.addParts(self.typesView.selectedIndexes())

    def addParts(self, indexes):
        """Adds the parts for the given indexes."""
        # add to current if that is a container type
        currentItem = self.scoreView.currentItem()
        for index in indexes:
            category = index.internalPointer()
            if category:
                part = category.items[index.row()]
                box = QGroupBox(self.partSettings)
                self.partSettings.addWidget(box)
                # determine the parent: current or root
                if currentItem and issubclass(part, currentItem.part.accepts()):
                    parent = currentItem
                    parent.setExpanded(True)
                else:
                    parent = self.scoreView
                item = PartItem(parent, part, box)
    
    def slotCurrentItemChanged(self, item):
        if isinstance(item, PartItem):
            self.partSettings.setCurrentWidget(item.box)
    
    def slotRemoveButtonClicked(self):
        self.scoreView.removeSelectedItems()
       
    def clear(self):
        """Called when the user clicks the clear button on this page."""
        self.scoreView.clear()

    def rootPartItem(self):
        """Returns the invisibleRootItem(), representing the tree of parts in the score view."""
        return self.scoreView.invisibleRootItem()
        

class PartItem(widgets.treewidget.TreeWidgetItem):
    """An item in the score tree widget."""
    def __init__(self, tree, part, box):
        """Initializes the item.
        
        tree: is the score tree widget,
        part: is the Part instance that creates the widgets
        box: the QGroupBox that is created for this item in the stacked widget.
        
        """
        super(PartItem, self).__init__(tree)
        self.part = part()
        self.box = box
        layout = QVBoxLayout()
        box.setLayout(layout)
        self.part.createWidgets(layout)
        layout.addStretch(1)
        app.translateUI(self)
        
        flags = (
            Qt.ItemIsSelectable |
            Qt.ItemIsDragEnabled |
            Qt.ItemIsEnabled
        )
        if issubclass(part, parts._base.Container):
            flags |= Qt.ItemIsDropEnabled
        self.setFlags(flags)
        
    def translateUI(self):
        self.setText(0, self.part.title())
        self.box.setTitle(self.part.title())
        self.part.translateWidgets()
        
    def cleanup(self):
        self.box.deleteLater()





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Cut selected text and assign it to a LilyPond variable.
"""

from __future__ import unicode_literals

from PyQt4.QtGui import QTextCursor

import cursortools
import tokeniter
import ly.lex.lilypond
import metainfo
import indent
import inputdialog


def cut_assign(cursor):
    """Cuts selected text and assigns it to a LilyPond variable."""
    # ask the variable name
    name = inputdialog.getText(None, _("Cut and Assign"), _(
        "Please enter the name for the variable to assign the selected "
        "text to:"), regexp="[A-Za-z]+")
    if not name:
        return
    
    cursortools.strip_selection(cursor)
    
    # determine state at cursor
    block = cursortools.block(cursor)
    state = tokeniter.state(block)
    for t in tokeniter.partition(cursor).left:
        state.follow(t)
    
    mode = ""
    for p in state.parsers():
        if isinstance(p, ly.lex.lilypond.ParseInputMode):
            if isinstance(p, ly.lex.lilypond.ParseLyricMode):
                mode = " \\lyricmode"
            elif isinstance(p, ly.lex.lilypond.ParseChordMode):
                mode = " \\chordmode"
            elif isinstance(p, ly.lex.lilypond.ParseFigureMode):
                mode = " \\figuremode"
            elif isinstance(p, ly.lex.lilypond.ParseDrumMode):
                mode = " \\drummode"
            break

    # find insertion place:
    found = False
    while block.previous().isValid():
        block = block.previous()
        state = tokeniter.state(block)
        if isinstance(state.parser(), ly.lex.lilypond.ParseGlobal):
            found = True
            break
        tokens = tokeniter.tokens(block)
        for t in tokens:
            if isinstance(t, ly.lex.lilypond.Name):
                found = True
                break
            elif not isinstance(t, (ly.lex.Space, ly.lex.Comment)):
                break
        if found:
            break
    insert = QTextCursor(block)
    text = cursor.selection().toPlainText()
    space = '\n' if '\n' in text else ' '
    text = ''.join((name, ' =', mode, ' {', space, text, space, '}\n\n'))
    with cursortools.compress_undo(cursor):
        cursor.insertText('\\' + name)
        pos = insert.selectionStart()
        insert.insertText(text)
    if metainfo.info(cursor.document()).auto_indent:
        insert.setPosition(pos, QTextCursor.KeepAnchor)
        with cursortools.compress_undo(insert, True):
            indent.re_indent(insert)





"""
This package brings the most important parts of the functionality of the
PortMIDI library in a flexible way to Python.

This module provides a simple API to control the most important parts of
PortMIDI, very much like the pygame.midi api.

It needs the PortMIDI library to be installed, and it is also recommended to
install the 'pypm' or 'pyportmidi' module providing a low-level access to the
PortMIDI library. This is a PyRex-based interface by John Harrison, which is
in some Linux distributions part of a package named 'python-portmidi' or
'python-pypm', and also part of the official PortMIDI package, but quite often
not installed by default.

The PyGame package as of version 1.9.1 also installs the 'pypm' module.

This module tries different ways to find the Python binding module. If that
fails, the PortMIDI library itself is loaded (if found) via ctypes.

To affect the order in which the _setup() function tries to import PortMIDI,
the global try_order list can be changed. The first name in the list is tried
first. This only works before init() is called for the first time.

This module can always be imported, but only init() and available() can be used
if PortMIDI itself is not available.

This module itself is in the public domain, and written by Wilbert Berendsen
in fall 2011.


Usage:

import portmidi

# test if portmidi really loaded:
if portmidi.available():
    ... enable MIDI stuff 

# start using PortMIDI: init and quit may be used multiple times
portmidi.init()


# list devices:
for i in range(portmidi.get_count()):
    print(i, portmidi.get_device_info(i))


# play notes to device number 2:
o = portmidi.Output(2, 50)
o.note_on(60)
o.note_off(60)


# play a scale:
time = 0
scale = []
for note in 60, 62, 64, 65, 67, 69, 71, 72:
    scale.append([[0x90, note, 80], time])
    scale.append([[0x80, note], time + 400])
    time += 500

o.write([[msg, portmidi.time() + time] for msg, time in scale])

Note that you always have to specify the time when a note is to be played.

"""

import atexit
import collections


__all__ = [
    'available', 'init', 'quit',
    'get_count', 'get_device_info',
    'get_default_input_id', 'get_default_output_id',
    'time',
    'Input', 'Output',
]

pypm = None
_initialized = None



# you can change this before calling init() for the first time
try_order = ['pypm', 'pyportmidi', 'pygame', 'ctypes']


def available():
    """Returns True if PortMIDI is available."""
    return _setup()

def init():
    """Initializes the PortMIDI library for use.
    
    It is safe to call this more than once.
    
    """
    global _initialized
    if _setup() and not _initialized:
        pypm.Initialize()
        _initialized = True

def quit():
    """Terminates the PortMIDI library.
    
    It is safe to call this more than once.
    On application exit this is also called.
    
    """
    global _initialized
    if pypm and _initialized:
        pypm.Terminate()
        _initialized = False

def get_count():
    """Returns the number if available MIDI devices."""
    _check_initialized()
    return pypm.CountDevices()

def get_default_input_id():
    """Returns the default input device number."""
    _check_initialized()
    return pypm.GetDefaultInputDeviceID()

def get_default_output_id():
    """Returns the default output device number."""
    _check_initialized()
    return pypm.GetDefaultOutputDeviceID()

def get_device_info(device_id):
    """Returns information about a midi device.
    
    A five-tuple is returned:
    (interf, name, isinput, isoutput, isopen).
    
    """
    _check_initialized()
    return device_info(*pypm.GetDeviceInfo(device_id))

def time():
    """Returns the current time in ms of the PortMIDI timer."""
    return pypm.Time()


class Input(object):
    """Reads MIDI input from a device."""
    def __init__(self, device_id, buffer_size=4096):
        self._input = None
        _check_initialized()
        _check_device_id(device_id)
        info = get_device_info(device_id)
        if not info.isinput:
            raise MidiException("not an input device")
        
        self._input = pypm.Input(device_id, buffer_size)

    def close(self):
        """Closes the input stream."""
        if self._input:
            self._input.Close()
        self._input = None

    def read(self, num_events):
        """reads num_events midi events from the buffer."""
        return self._input.Read(num_events)

    def poll(self):
        """Returns True if there's data, otherwise False."""
        r = self._input.Poll()
        if r == pypm.TRUE:
            return True
        elif r == pypm.FALSE:
            return False
        else:
            raise MidiException(pypm.GetErrorText(r).decode('utf-8'))


class Output(object):
    """Writes MIDI output to a device."""
    def __init__(self, device_id, latency = 0, buffer_size = 4096):
        self._output = None
        _check_initialized()
        _check_device_id(device_id)
        info = get_device_info(device_id)
        if not info.isoutput:
            raise MidiException("not an output device")
        
        self._output = pypm.Output(device_id, latency)

    def close(self):
        """Closes the output stream."""
        if self._output:
            self._output.Close()
        self._output = None

    def write(self, data):
        """Writes a list of MIDI data to the output.
        
        Each element of the list should be a list[message, timestamp].
        Each message is again a list: [status, data, data, ...]
        
        """
        self._output.Write(data)

    def write_short(self, status, data1 = 0, data2 = 0):
        """Output MIDI information of 3 bytes or less."""
        self._output.WriteShort(status, data1, data2)

    def write_sys_ex(self, timestamp, message):
        """Writes a timestamped System-Exclusive MIDI message.
        
        The message may be a list of integers or a bytes string.
        
        """
        self._output.WriteSysEx(timestamp, message)

    def note_on(self, note, velocity=80, channel = 0):
        """Turns a midi note on."""
        _check_channel(channel)
        self.write_short(0x90 + channel, note, velocity)

    def note_off(self, note, velocity=0, channel = 0):
        """Turns a midi note off."""
        _check_channel(channel)
        self.write_short(0x80 + channel, note, velocity)

    def set_instrument(self, instrument_id, channel = 0):
        """Select an instrument."""
        if not 0 <= instrument_id <= 127:
            raise ValueError("invalid instrument id")
        _check_channel(channel)
        self.write_short(0xC0 + channel, instrument_id)


class MidiException(Exception):
    """Raised on MIDI-specific errors."""
    pass



# helper functions

device_info = collections.namedtuple('device_info',
    'interf name isinput isoutput isopen')

def _check_device_id(device_id):
    if not 0 <= device_id < get_count():
        raise ValueError("invalid device id")

def _check_channel(channel):
    if not 0 <= channel <= 15:
        raise ValueError("invalid channel number (must be 0..15)")

def _check_initialized():
    if not _initialized:
        raise RuntimeError("PortMIDI not initialized.")

def _setup():
    """Tries to import PortMIDI in the order given in the try_order global.
    
    Only one time the import is tried.
    Returns True if PortMIDI could be loaded.
    
    """
    global pypm
    if pypm is None:
        for name in try_order:
            try:
                pypm = globals()['_do_import_' + name]()
                break
            except ImportError:
                continue
        else:
            pypm = False
    return bool(pypm)

def _load_module(name):
    """Loads and returns a single module.
    
    The name may be dotted, but no parent modules are imported.
    Raises ImportError when the module can't be found.
    
    """
    import imp
    path = None
    for n in name.split('.'):
        file_handle, path, desc = imp.find_module(n, path and [path])
    return imp.load_module(n, file_handle, path, desc)


# these functions try to import PortMIDI, returning the module
def _do_import_pypm():
    """Tries to import pypm (the c binding module) directly.
    
    Newer Ubuntu releases provide a python-pypm package with this module.
    
    """
    import pypm
    # Reject incompatible API, such as ActiveState PyPM
    if not hasattr(pypm, "CountDevices"):
        del pypm
        raise ImportError("Unsupported pypm API")
    return pypm

def _do_import_pyportmidi():
    """This tries to import the c Python module from the PortMIDI distribution.
    
    Unfortunately not many Linux distros also install the Python binding.
    
    """
    return _load_module('pyportmidi._pyportmidi')

def _do_import_pygame():
    """This tries to import the pypm module from the pygame package."""
    # We don't use 'from pygame import pypm' as that also imports a large
    # part of pygame which we probably don't use.
    return _load_module('pygame.pypm')

def _do_import_ctypes():
    """This tries to load PortMIDI via ctypes."""
    from . import ctypes_pypm
    return ctypes_pypm





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
This is only imported on Mac OS X, from main.py.

It initializes various stuff that's only relevant on Mac OS X.

"""

from __future__ import unicode_literals

from PyQt4.QtCore import QTimer

import app

@app.mainwindowClosed.connect
def check_open_window():
    if not app.windows:
        from . import globalmenu
        QTimer.singleShot(0, globalmenu.setup)

def initialize():
    # on Mac OS X, handle FileOpen requests (e.g. double-clicking a file in the
    # Finder), these events also can occur right on application start.
    # We do this just before creating the window, so that when multiple files
    # are opened on startup (I don't know whether that really could happen),
    # they are not made the current document, as that slows down loading
    # multiple documents drastically.
    from . import file_open_eventhandler
    file_open_eventhandler.initialize()
    
    # handle window icon drag events
    from . import icon_drag_eventhandler
    icon_drag_eventhandler.initialize()
    
    # on mac os, the app should remain running, even if there is no main window 
    # anymore. In this case, we setup a basic global menu.
    app.qApp.setQuitOnLastWindowClosed(False)




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Setup the application language.

Also contains a function to get language preferences from the operation
system.

"""

import locale

from PyQt4.QtCore import QLocale, QSettings, QTimer

import app

from . import find, install, available

__all__ = ['preferred', 'current', 'default']


_currentlanguage = None


def preferred():
    """Return a list of language codes from the operating system preferences.
    
    Language- and country codes will always be separated with an underscore '_'.
    
    """
    try:
        langs = QLocale().uiLanguages()
    except AttributeError:
        # QLocale.uiLanguages is not in Qt 4.7 (only Qt4.8+)
        langs = []
    else:
        # in some systems, language/country codes have '-' and not '_'
        langs = [l.replace('-', '_') for l in langs]
    if not langs:
        try: 
            langs.append(locale.getdefaultlocale()[0])
        except ValueError:
            pass
    return langs

def default():
    """Return the first preferred system default UI language that is available in Frescobaldi.
    
    May return None, if none of the system preferred languages is available
    in Frescobaldi.
    
    """
    av_langs = available()
    av_langs.append("en")
    for lang in preferred():
        if lang in av_langs or lang.split('_')[0] in av_langs:
            return lang

def current():
    """Returns the currently active UI language code.
    
    A name is always returned, which can be "C", meaning no translation
    is desired.
    
    """
    return QSettings().value("language", "", type("")) or default() or "C"

def _setup():
    """Set application language according to settings."""
    global _currentlanguage
    language = current()
    if _currentlanguage is not None and language != _currentlanguage:
        QTimer.singleShot(0, app.languageChanged)
    _currentlanguage = language
    if language != "C":
        mo = find(language)
        if mo:
            install(mo)
            return
    install(None)

@app.oninit
def _start_up():
    """Initialize GUI translations. Called op app startup."""
    from . import qtranslator
    # XXX TEMP qtranslator.initialize() segfaults on my system in Python3
    import sys
    if sys.version_info < (3, 0):
        qtranslator.initialize()
    _setup()


app.settingsChanged.connect(_setup)




"""
In this directory you can place hyphenation dictionaries (with names like
"hyph_nl_NL.dic", e.g. from OpenOffice.org or MySpell), which will then be found
by the Frescobaldi hyphenation dialog (see lyrics.py, hyphendialog.py and
hyphenator.py).
"""

import os

path = os.path.abspath(__path__[0])

del os



# Python midifile package -- parse, load and play MIDI files.
# Copyright (c) 2011 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
midifile.parser -- parses MIDI file data.

This is a simple module that can parse data from a MIDI file and
its tracks.

A basic event factory returns the MIDI events as simple named tuples,
but you can subclass the event factory for more sophisticated behaviour.

Runs with Python 2.6, 2.7 and 3.

"""

from __future__ import unicode_literals
from __future__ import print_function

import sys
import struct

from . import event


PY2 = sys.version_info[0] == 2


unpack_midi_header = struct.Struct(b'>hhh').unpack
unpack_int = struct.Struct(b'>i').unpack


def get_chunks(s):
    """Splits a MIDI file bytes string into chunks.
    
    Yields (b'Name', b'data') tuples.
    
    """
    pos = 0
    while pos < len(s):
        name = s[pos:pos+4]
        size, = unpack_int(s[pos+4:pos+8])
        yield name, s[pos+8:pos+8+size]
        pos += size + 8

    
def parse_midi_data(s):
    """Parses MIDI file data from the bytes string s.
    
    Returns a three tuple (format_type, time_division, tracks).
    Every track is an unparsed bytes string.
    
    May raise ValueError or IndexError in case of invalid MIDI data.
    
    """
    chunks = get_chunks(s)
    for name, data in chunks:
        if name == b'MThd':
            fmt, ntracks, division = unpack_midi_header(data[:6])
            tracks = [data for name, data in chunks if name == b'MTrk']
            return fmt, division, tracks
        break
    raise ValueError("invalid midi data")


def read_var_len(s, pos):
    """Reads variable-length integer from byte string s starting on pos.
    
    Returns the value and the new position.
    
    """
    value = 0
    while True:
        i = s[pos]
        pos += 1
        value = value * 128 + (i & 0x7F)
        if not i & 0x80:
            return value, pos


def parse_midi_events(s, factory=None):
    """Parses the bytes string s (typically a track) for MIDI events.
    
    If factory is given, it should be an EventFactory instance that
    returns objects describing the event.
    
    Yields two-tuples (delta, event).
    
    Raises ValueError or IndexError on invalid MIDI data.
    
    """
    if factory is None:
        factory = event.EventFactory()
        
    running_status = None
    
    if PY2:
        s = bytearray(s)
    
    pos = 0
    while pos < len(s):
        
        delta, pos = read_var_len(s, pos)
        
        status = s[pos]
        if status & 0x80:
            running_status = status
            pos += 1
        elif not running_status:
            raise ValueError("invalid running status")
        else:
            status = running_status
        
        ev_type = status >> 4
        channel = status & 0x0F
        
        if ev_type <= 0x0A:
            # note on, off or aftertouch
            note = s[pos]
            value = s[pos+1]
            pos += 2
            ev = factory.note_event(ev_type, channel, note, value)
        elif ev_type >= 0x0F:
            running_status = None
            if status == 0xFF:
                # meta event
                meta_type = s[pos]
                meta_size, pos = read_var_len(s, pos+1)
                meta_data = s[pos:pos+meta_size]
                pos += meta_size
                ev = factory.meta_event(meta_type, meta_data)
            else:
                # some sort of sysex
                sysex_size, pos = read_var_len(s, pos)
                sysex_data = s[pos:pos+sysex_size]
                pos += sysex_size
                ev = factory.sysex_event(status, sysex_data)
        elif ev_type == 0x0E:
            # Pitch Bend
            value = s[pos] + s[pos+1] * 128
            pos += 2
            ev = factory.pitchbend_event(channel, value)
        elif ev_type == 0xD:
            # Channel AfterTouch
            value = s[pos]
            pos += 1
            ev = factory.channelaftertouch_event(channel, value)
        elif ev_type == 0xB:
            # Controller
            number = s[pos]
            value = s[pos+1]
            pos += 2
            ev = factory.controller_event(channel, number, value)
        else: # ev_type == 0xC
            # Program Change
            number = s[pos]
            pos += 1
            ev = factory.programchange_event(channel, number)
        yield delta, ev


def time_events(track, time=0):
    """Yields two-tuples (time, event).
    
    The track is the generator returned by parse_midi_events,
    the time is accumulated from the given starting time (defaulting to 0).
    
    """
    for delta, ev in track:
        time += delta
        yield time, ev


def time_events_grouped(track, time=0):
    """Yields two-tuples (time, event_list).
    
    Every event_list is a Python list of all events happening on that time.
    The track is the generator returned by parse_midi_events,
    the time is accumulated from the given starting time (defaulting to 0).
    
    """
    evs = []
    for delta, ev in track:
        if delta:
            if evs:
                yield time, evs
                evs = []
            time += delta
        evs.append(ev)
    if evs:
        yield time, evs



if __name__ == '__main__':
    """Test specified MIDI files."""
    import sys
    files = sys.argv[1:]
    for f in files:
        with open(f, 'rb') as midifile:
            s = midifile.read()
        ftm, div, tracks = parse_midi_data(s)
        try:
            for t in tracks:
                print(list(parse_midi_events(t)))
        except Exception as e:
            print('error in:', f)
            print(e)






# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Percussion part types.
"""

from __future__ import unicode_literals

from PyQt4.QtGui import QCheckBox, QComboBox, QGridLayout, QHBoxLayout, QLabel, QSpinBox

import listmodel
import ly.dom

from . import _base
from . import register



class PitchedPercussionPart(_base.SingleVoicePart):
    """Base class for pitched percussion types."""
    
    
class Timpani(PitchedPercussionPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Timpani")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Timpani", "Tmp.")
        
    midiInstrument = 'timpani'
    clef = 'bass'
    octave = -1


class Xylophone(PitchedPercussionPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Xylophone")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Xylophone", "Xyl.")
        
    midiInstrument = 'xylophone'


class Marimba(_base.PianoStaffPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Marimba")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Marimba", "Mar.")
    
    midiInstrument = 'marimba'

    def createWidgets(self, layout):
        super(Marimba, self).createWidgets(layout)
        self.lowerVoices.setMinimum(0)
    
    def translateWidgets(self):
        self.upperVoicesLabel.setText(_("Upper staff:"))
        self.lowerVoicesLabel.setText(_("Lower staff:"))
        self.lowerVoices.setToolTip(_(
            "Set the number of voices to 0 to disable the second staff."))

    def build(self, data, builder):
        if self.lowerVoices.value():
            super(Marimba, self).build(data, builder)
        else:
            data.nodes.append(self.buildStaff(data, builder, None, 1))


class Vibraphone(Marimba):
    @staticmethod
    def title(_=_base.translate):
        return _("Vibraphone")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Vibraphone", "Vib.")
        
    midiInstrument = 'vibraphone'


class TubularBells(PitchedPercussionPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Tubular bells")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Tubular bells", "Tub.")
        
    midiInstrument = 'tubular bells'


class Glockenspiel(PitchedPercussionPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Glockenspiel")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Glockenspiel", "Gls.")
        
    midiInstrument = 'glockenspiel'


class Carillon(_base.PianoStaffPart):
    @staticmethod
    def title(_=_base.translate):
        return _("Carillon")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Carillon", "Car.")
    
    midiInstrument = 'tubular bells' # anyone knows better?
    
    def translateWidgets(self):
        super(Carillon, self).translateWidgets()
        self.upperVoicesLabel.setText(_("Manual staff:"))
        self.lowerVoicesLabel.setText(_("Pedal staff:"))
    
    def build(self, data, builder):
        p = ly.dom.PianoStaff()
        builder.setInstrumentNamesFromPart(p, self, data)
        s = ly.dom.Sim(p)
        # add two staves, with a respective number of voices.
        self.buildStaff(data, builder, 'manual', 1, self.upperVoices.value(), s)
        self.buildStaff(data, builder, 'pedal', 0, self.lowerVoices.value(), s, "bass")
        data.nodes.append(p)


class Drums(_base.Part):
    @staticmethod
    def title(_=_base.translate):
        return _("Drums")
    
    @staticmethod
    def short(_=_base.translate):
        return _("abbreviation for Drums", "Dr.")
        
    def createWidgets(self, layout):
        self.voicesLabel = QLabel()
        self.voices = QSpinBox(minimum=1, maximum=4, value=1)
        self.drumStyleLabel = QLabel()
        self.drumStyle = QComboBox()
        self.drumStyle.setModel(listmodel.ListModel(drumStyles, self.drumStyle, display=listmodel.translate))
        self.drumStems = QCheckBox()
        
        box = QHBoxLayout()
        box.addWidget(self.voicesLabel)
        box.addWidget(self.voices)
        layout.addLayout(box)
        box = QHBoxLayout()
        box.addWidget(self.drumStyleLabel)
        box.addWidget(self.drumStyle)
        layout.addLayout(box)
        layout.addWidget(self.drumStems)
        
    def translateWidgets(self):
        self.voicesLabel.setText(_("Voices:"))
        self.drumStyleLabel.setText(_("Style:"))
        self.drumStems.setText(_("Remove stems"))
        self.drumStems.setToolTip(_("Remove the stems from the drum notes."))
        self.drumStyle.model().update()
    
    def assignDrums(self, data, name = None):
        r"""Creates an empty name = \drummode assignment.

        Returns the assignment.
        
        """
        a = data.assign(name)
        s = ly.dom.DrumMode(a)
        ly.dom.Identifier(data.globalName, s)
        ly.dom.LineComment(_("Drums follow here."), s)
        ly.dom.BlankLine(s)
        return a

    def build(self, data, builder):
        p = ly.dom.DrumStaff()
        s = ly.dom.Simr(p)
        if self.voices.value() > 1:
            for i in range(1, self.voices.value() + 1):
                q = ly.dom.Seq(ly.dom.DrumVoice(parent=s))
                ly.dom.Text('\\voice' + ly.util.int2text(i), q)
                a = self.assignDrums(data, 'drum' + ly.util.int2text(i))
                ly.dom.Identifier(a.name, q)
        else:
            a = self.assignDrums(data, 'drum')
            ly.dom.Identifier(a.name, s)
        builder.setInstrumentNamesFromPart(p, self, data)
        i = self.drumStyle.currentIndex()
        if i > 0:
            v = ('drums', 'timbales', 'congas', 'bongos', 'percussion')[i]
            p.getWith()['drumStyleTable'] = ly.dom.Scheme(v + '-style')
            v = (5, 2, 2, 2, 1)[i]
            ly.dom.Line("\\override StaffSymbol #'line-count = #{0}".format(v), p.getWith())
        if self.drumStems.isChecked():
            ly.dom.Line("\\override Stem #'stencil = ##f", p.getWith())
            ly.dom.Line("\\override Stem #'length = #3  % " + _("keep some distance."),
                p.getWith())
        data.nodes.append(p)


drumStyles = (
    lambda: _("Drums (5 lines, default)"),
    lambda: _("Timbales-style (2 lines)"),
    lambda: _("Congas-style (2 lines)"),
    lambda: _("Bongos-style (2 lines)"),
    lambda: _("Percussion-style (1 line)"),
)
    


register(
    lambda: _("Percussion"),
    [
        Timpani,
        Xylophone,
        Marimba,
        Vibraphone,
        TubularBells,
        Glockenspiel,
        Carillon,
        Drums,
    ])



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The Quick Insert panel widget.
"""

from __future__ import unicode_literals

import weakref

from PyQt4.QtCore import QSettings
from PyQt4.QtGui import (QComboBox, QHBoxLayout, QLabel, QToolBox, QToolButton,
                         QVBoxLayout, QWidget)

import app
import userguide.util
import icons
import symbols
import widgets.toolboxwheeler

from . import articulations
from . import barlines
from . import dynamics
from . import spanners


class QuickInsert(QWidget):
    def __init__(self, dockwidget):
        super(QuickInsert, self).__init__(dockwidget)
        self._dockwidget = weakref.ref(dockwidget)
        # filled in by ButtonGroup subclasses
        self.actionDict = {}
        
        layout = QVBoxLayout()
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)
        
        self.helpButton = QToolButton(
            icon = icons.get("help-contents"),
            autoRaise = True,
            clicked = lambda: userguide.show("quickinsert"))
        self.directionLabel = QLabel()
        self.direction = QComboBox()
        self.direction.addItems(['', '', ''])
        self.direction.setItemIcon(0, icons.get("go-up"))
        self.direction.setItemIcon(2, icons.get("go-down"))
        self.direction.setCurrentIndex(1)
        hor = QHBoxLayout()
        hor.setContentsMargins(0, 0, 0, 0)
        hor.addWidget(self.helpButton)
        hor.addWidget(self.directionLabel)
        hor.addWidget(self.direction)
        layout.addLayout(hor)
        
        self.toolbox = QToolBox(self)
        widgets.toolboxwheeler.ToolBoxWheeler(self.toolbox)
        layout.addWidget(self.toolbox)
        
        for cls in (
                articulations.Articulations,
                dynamics.Dynamics,
                spanners.Spanners,
                barlines.BarLines,
            ):
            widget = cls(self)
            self.toolbox.addItem(widget, widget.icon(), '')
        
        app.translateUI(self)
        userguide.openWhatsThis(self)
        
        # restore remembered current page
        name = QSettings().value("quickinsert/current_tool", "", type(""))
        if name:
            for i in range(self.toolbox.count()):
                if name == self.toolbox.widget(i).__class__.__name__.lower():
                    self.toolbox.setCurrentIndex(i)
                    break
        self.toolbox.currentChanged.connect(self.slotCurrentChanged)
        
    def slotCurrentChanged(self, index):
        name = self.toolbox.widget(index).__class__.__name__.lower()
        QSettings().setValue("quickinsert/current_tool", name)
    
    def translateUI(self):
        self.setWhatsThis(_(
            "<p>With the Quick Insert Panel you can add various music "
            "elements to the current note or selected music.</p>\n"
            "<p>See {link} for more information.</p>").format(link=
                userguide.util.format_link("quickinsert")))
        self.helpButton.setToolTip(_("Help"))
        self.directionLabel.setText(_("Direction:"))
        for item, text in enumerate((_("Up"), _("Neutral"), _("Down"))):
            self.direction.setItemText(item, text)
        for i in range(self.toolbox.count()):
            self.toolbox.setItemText(i, self.toolbox.widget(i).title())
            self.toolbox.setItemToolTip(i, self.toolbox.widget(i).tooltip())
            
    def actionForName(self, name):
        """This is called by the ShortcutCollection of our dockwidget, e.g. if the user presses a key."""
        try:
            return self.actionDict[name]
        except KeyError:
            pass

    def dockwidget(self):
        return self._dockwidget()





# process.py -- A very simple wrapper around QProcess
#
# Copyright (c) 2012 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A very simple wrapper around QProcess, and a scheduler to enable running
one process at a time.
"""

__all__ = ['Process', 'Scheduler']

from PyQt4.QtCore import QObject, QProcess, pyqtSignal


class Process(QObject):
    """A very simple wrapper to run a QProcess.
    
    Initialize the process with the command line (as a list) to run.
    Then (or later) call start() to run it. At that moment the 'process'
    attribute is available, which contains the QProcess instance managing
    the process.
    
    The done() signal is emitted with a boolean success value. In slots
    connected (synchronously) the process attribute is still available.
    
    """
    
    done = pyqtSignal(bool)
    
    def __init__(self, command):
        """Sets up the command to run."""
        QObject.__init__(self)
        self.command = command
    
    def start(self):
        """Really starts a QProcess, executing the command line."""
        self.setup()
        self.process.start(self.command[0], self.command[1:])
    
    def setup(self):
        """Called on start(), sets up the QProcess in the process attribute."""
        self.process = p = QProcess()
        p.finished.connect(self._finished)
        p.error.connect(self._error)
        
    def _finished(self, exitCode):
        self._done(exitCode == 0)
    
    def _error(self):
        self._done(False)
    
    def _done(self, success):
        self.done.emit(success)
        self.cleanup()
    
    def cleanup(self):
        """Deletes the process."""
        self.process.deleteLater()
        del self.process


class Scheduler(object):
    """A very simple scheduler that runs one Process at a time.
    
    You can use this to run e.g. commandline tools asynchronously and you
    don't want to have them running at the same time.
    
    """
    def __init__(self):
        self._schedule = []
    
    def add(self, process):
        """Adds the process to run."""
        process.done.connect(self._done)
        self._schedule.append(process)
        if len(self._schedule) == 1:
            self._schedule[0].start()
    
    def remove(self, process):
        """Removes the process from the schedule.
        
        This only works if the process has not been started yet.
        
        """
        if process in self._schedule[1:]:
            self._schedule.remove(process)
            process.done.disconnect(self._done)
    
    def _done(self):
        del self._schedule[0]
        if self._schedule:
            self._schedule[0].start()





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
A QNetworkAccessManager subclass with easy registration of custom url schemes.
"""

from __future__ import unicode_literals

from PyQt4.QtCore import QThread, QTimer, QUrl
from PyQt4.QtNetwork import QNetworkAccessManager, QNetworkReply, QNetworkRequest


class NetworkAccessManager(QNetworkAccessManager):
    """A QNetworkAccessManager subclass with easy registration of custom url schemes.
    
    Use the registerHandler() method to add custom scheme handlers.
    The registerHtmlHandler() method is used to add a simple callable as handler that
    gets a QUrl and should return the HTML as a normal string.
    
    The headers instance attribute is a dictionary (empty by default) containing
    extra headers (as key, value) that are added to outgoing requests.
    
    """
    
    def __init__(self, parent=None):
        QNetworkAccessManager.__init__(self, parent)
        self._dispatcher = {}
        self.headers = {}
        
    def createRequest(self, operation, request, data):
        try:
            requestFunc = self._dispatcher[request.url().scheme()]
        except KeyError:
            self.addHeadersToRequest(request)
            return QNetworkAccessManager.createRequest(self, operation, request, data)
        return requestFunc(self, operation, request, data)
    
    def addHeadersToRequest(self, request):
        """Called on outgoing requests and should add raw headers to the request.
        
        The default implementation of this method simply adds all the headers in the
        headers instance attribute.
        
        """
        for name, value in self.headers.items():
            request.setRawHeader(name, value)
    
    def registerHandler(self, scheme, handler):
        """Registers a handler for the given scheme.
        
        The handler is called with four arguments (manager, operation, request,
        data), just like the QNetworkAccessManager.createRequest method, and
        should return a QNetworkReply instance.
        
        """
        self._dispatcher[scheme] = handler
        
    def registerHtmlHandler(self, scheme, handler, threaded=False, encoding="UTF-8"):
        """Registers a simple callable as the handler for the given scheme.
        
        The handler only gets a GET URL (QUrl) and should return a HTML string.
        If threaded is True, the handler is called in a background thread.
        The encoding defaults to UTF-8 and is used for encoding the HTML returned
        by then handler and also set in the Content-Type header.
        
        """
        cls = ThreadedHtmlReply if threaded else HtmlReply
        def createRequest(mgr, operation, request, data):
            return cls(mgr, request.url(), handler, encoding)
        self._dispatcher[scheme] = createRequest

    def unregisterHandler(self, scheme):
        """Removes the special handling for the given scheme."""
        try:
            del self._dispatcher[scheme]
        except KeyError:
            pass


class HtmlReplyBase(QNetworkReply):
    """Abstract base class for a QNetworkReply that represents a generated HTML string."""
    def __init__(self, manager, url, handler, encoding="UTF-8"):
        QNetworkReply.__init__(self, manager)
        self.setUrl(url)
        self._handler = handler
        self._encoding = encoding
    
    def callHandler(self):
        self._content = self._handler(self.url()).encode(self._encoding)
        
    def outputReady(self):
        self._offset = 0
        self.setHeader(QNetworkRequest.ContentTypeHeader, "text/html; charset={0}".format(self._encoding))
        self.setHeader(QNetworkRequest.ContentLengthHeader, len(self._content))
        self.open(QNetworkReply.ReadOnly | QNetworkReply.Unbuffered)
    
    def emitSignals(self):
        self.readyRead.emit()
        self.finished.emit()
    
    def abort(self):
        pass
    
    def bytesAvailable(self):
        return len(self._content) - self._offset
    
    def isSequential(self):
        return True
    
    def readData(self, maxSize):
        if self._offset < len(self._content):
            end = min(self._offset + maxSize, len(self._content))
            data = self._content[self._offset:end]
            self._offset = end
            return data


class HtmlReply(HtmlReplyBase):
    """QNetworkReply that generates a HTML string by calling handler(url).encode(encoding)."""
    def __init__(self, manager, url, handler, encoding="UTF-8"):
        HtmlReplyBase.__init__(self, manager, url, handler, encoding)
        self.callHandler()
        self.outputReady()
        QTimer.singleShot(0, self.emitSignals)


class ThreadedHtmlReply(HtmlReplyBase):
    """HtmlReply that calls handler(url) in a background thread."""
    def __init__(self, manager, url, handler, encoding="UTF-8"):
        HtmlReplyBase.__init__(self, manager, url, handler, encoding)
        self._thread = Thread(self.callHandler)
        self._thread.finished.connect(self.threadFinished)
        self._thread.start()
    
    def threadFinished(self):
        self.outputReady()
        self.emitSignals()


class Thread(QThread):
    """QThread that runs a single callable."""
    def __init__(self, func):
        QThread.__init__(self)
        self._func = func
    
    def run(self):
        self._func()





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The Music View's Edit in Place dialog.
"""

from __future__ import unicode_literals

from PyQt4.QtCore import QSettings, QSize
from PyQt4.QtGui import (QAction, QKeySequence, QPlainTextDocumentLayout,
                         QPlainTextEdit, QTextCharFormat, QTextCursor,
                         QTextDocument)

import app
import actioncollectionmanager
import autocomplete.analyzer
import autocomplete.completer
import cursordiff
import cursortools
import userguide
import highlighter
import homekey
import indent
import matcher
import metainfo
import qutil
import textformats
import tokeniter
import widgets.dialog

from . import tooltip


def edit(parent, cursor, position=None):
    dlg = Dialog(parent)
    dlg.finished.connect(dlg.deleteLater)
    dlg.edit(cursor)
    dlg.popup(position)


class Dialog(widgets.dialog.Dialog):
    """Dialog containing a short text edit field to edit one line."""
    def __init__(self, parent=None):
        super(Dialog, self).__init__(parent)
        self._document = None
        self.messageLabel().setWordWrap(True)
        self.document = d = QTextDocument()
        d.setDocumentLayout(QPlainTextDocumentLayout(d))
        self.highlighter = highlighter.highlighter(d)
        self.view = View(d)
        self.matcher = Matcher(self.view)
        self.completer = Completer(self.view)
        self.setMainWidget(self.view)
        userguide.addButton(self.buttonBox(), "musicview_editinplace")
        # action for completion popup
        self._showPopupAction = QAction(None, triggered=self.slotCompletionPopup)
        self.addAction(self._showPopupAction)
        # make Ctrl+Return accept the dialog
        self.button("ok").setShortcut(QKeySequence("Ctrl+Return"))
        qutil.saveDialogSize(self, "musicview/editinplace/dialog/size")
        
        self.accepted.connect(self.save)
        app.translateUI(self)
        app.settingsChanged.connect(self.readSettings)
        self.readSettings()
        
    def translateUI(self):
        self.setWindowTitle(app.caption(_("Edit in Place")))
        self.updateMessage()
    
    def readSettings(self):
        self._showPopupAction.setShortcut(
            actioncollectionmanager.action("autocomplete", "popup_completions").shortcut())
        
    def edit(self, cursor):
        """Edit the block at the specified QTextCursor."""
        if self._document:
            self._document.closed.disconnect(self.reject)
        self._document = cursor.document()
        self._document.closed.connect(self.reject)
        
        # don't change the cursor
        c = self._range = QTextCursor(cursor)
        cursorpos = c.position() - c.block().position()
        cursortools.strip_indent(c)
        indentpos = c.position() - c.block().position()
        c.movePosition(QTextCursor.EndOfBlock, QTextCursor.KeepAnchor)
        self.view.setPlainText(c.selection().toPlainText())
        
        self.highlighter.setInitialState(tokeniter.state(cursortools.block(cursor)))
        self.highlighter.setHighlighting(metainfo.info(cursor.document()).highlighting)
        self.highlighter.rehighlight()
        
        # let autocomplete query the real document as if we're at the start
        # of the current block
        self.completer.document_cursor = QTextCursor(cursor.block())
        self.completer.autoComplete = QSettings().value(
                        "autocomplete", True) not in ('false', False)
        
        cursor = self.view.textCursor()
        cursor.setPosition(max(0, cursorpos-indentpos))
        self.view.setTextCursor(cursor)
        
        self.updateMessage()
        
    def popup(self, position):
        """Show the dialog at the specified global QPoint."""
        geom = self.geometry()
        geom.moveCenter(position)
        if position.y() <= geom.height() + 60:
            geom.moveTop(position.y() + 60)
        else:
            geom.moveBottom(position.y() - 60)
        self.setGeometry(geom)
        self.view.setFocus()
        self.show()
    
    def save(self):
        """Called to perform the edits in the document."""
        cursor = QTextCursor(self._range)
        start = cursor.selectionStart()
        # use cursordiff; don't destroy point and click positions
        cursordiff.insert_text(cursor, self.view.toPlainText())
        cursor.setPosition(start, QTextCursor.KeepAnchor)
        with cursortools.compress_undo(cursor, True):
            # re-indent the inserted line(s)
            indent.re_indent(cursor)
        
    def updateMessage(self):
        """Called when a new cursor is set to edit, updates the message text."""
        if self._document:
            self.setMessage(
              _("Editing line {linenum} of \"{document}\" ({variable})").format(
                linenum = self._range.block().blockNumber() + 1,
                document = self._document.documentName(),
                variable = tooltip.get_definition(self._range) or _("<unknown>"),
            ))
        else:
            self.setMessage("<no document set>") # should never appear
    
    def slotCompletionPopup(self):
        self.completer.showCompletionPopup()


class View(QPlainTextEdit):
    """The text edit in the "Edit in Place" dialog."""
    def __init__(self, document):
        super(View, self).__init__()
        self.setDocument(document)
        self.setLineWrapMode(QPlainTextEdit.NoWrap)
        self.setCursorWidth(2)
        app.settingsChanged.connect(self.readSettings)
        self.readSettings()
    
    def readSettings(self):
        data = textformats.formatData('editor')
        self.setFont(data.font)
        self.setPalette(data.palette())
    
    def sizeHint(self):
        metrics = self.fontMetrics()
        return QSize(80 * metrics.width(" "),3 * metrics.height())
    
    def event(self, ev):
        """Reimplemented to avoid typing the line separator."""
        if ev == QKeySequence.InsertLineSeparator:
            return False
        return super(View, self).event(ev)


class Matcher(matcher.AbstractMatcher):
    """Looks for matches if the cursor moves."""
    def __init__(self, view):
        super(Matcher, self).__init__(view)
        self._highlighter = MatchHighlighter(view)
    
    def highlighter(self):
        return self._highlighter


class MatchHighlighter(widgets.arbitraryhighlighter.ArbitraryHighlighter):
    """Highlights the matches like { } or << >>."""
    def __init__(self, edit):
        super(MatchHighlighter, self).__init__(edit)
        app.settingsChanged.connect(self.readSettings)
        self.readSettings()
    
    def readSettings(self):
        self._baseColors = textformats.formatData('editor').baseColors
        self.reload()
    
    def textFormat(self, name):
        f = QTextCharFormat()
        f.setBackground(self._baseColors[name])
        return f


class Completer(autocomplete.completer.Completer):
    """A Completer providing completions for the Edit in Place popup.
    
    It can request information from the document specified by the
    document_cursor which can be set as an instance attribute.
    
    """
    document_cursor = None
    def __init__(self, view):
        super(Completer, self).__init__()
        self.setWidget(view)
    
    def analyzer(self):
        return Analyzer(self.document_cursor)


class Analyzer(autocomplete.analyzer.Analyzer):
    """An Analyzer looking at the line of text in the Edit in Place popup.
    
    It takes the document_cursor attribute on init from the Completer,
    so that the document the Edit in Place popup belongs to can be queried
    for information like defined variables, etc.
    
    """
    def __init__(self, cursor):
        self._document_cursor = cursor
    
    def document_cursor(self):
        """Reimplemented to return the cursor of the real document."""
        return self._document_cursor






# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Creates the commandline or Job to engrave a music document.
"""

from __future__ import unicode_literals

import os, sys

from PyQt4.QtCore import QSettings

import job
import documentinfo
import lilypondinfo


def info(document):
    """Returns a LilyPondInfo instance that should be used by default to engrave the document."""
    version = documentinfo.docinfo(document).version()
    if version and QSettings().value("lilypond_settings/autoversion", False, bool):
        return lilypondinfo.suitable(version)
    return lilypondinfo.preferred()


def defaultJob(document, args=None):
    """Return a default job for the document.
    
    The 'args' argument, if given, must be a list of commandline arguments
    that are given to LilyPond, and may enable specific preview modes.
    
    If args is not given, the Job will cause LilyPond to run in Publish mode,
    with point and click turned off.
    
    """
    filename, includepath = documentinfo.info(document).jobinfo(True)
    
    i = info(document)
    j = job.Job()
    
    command = [i.abscommand() or i.command]
    s = QSettings()
    s.beginGroup("lilypond_settings")
    if s.value("delete_intermediate_files", True, bool):
        command.append('-ddelete-intermediate-files')
    else:
        command.append('-dno-delete-intermediate-files')
    
    if args:
        command.extend(args)
    else:
        command.append('-dno-point-and-click')
    
    if s.value("default_output_target", "pdf", type("")) == "svg":
        command.append('-dbackend=svg')
    else:
        command.append('--pdf')
        
    command.extend('-I' + path for path in includepath)
    j.directory = os.path.dirname(filename)
    command.append(filename)
    j.command = command
    if s.value("no_translation", False, bool):
        j.environment['LANG'] = 'C'
    j.set_title("{0} {1} [{2}]".format(
        os.path.basename(i.command), i.versionString(), document.documentName()))
    return j



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Helper application stuff.
"""

from __future__ import unicode_literals

import os
import re
import subprocess
import sys

from PyQt4.QtCore import QSettings
from PyQt4.QtGui import QDesktopServices, QMessageBox

def shell_split(cmd):
    """Split a string like the UNIX shell, returning a list.
    
    Double quoted parts are kept together with the quotes removed.
    
    """
    # remove double quotes, keeping quoted parts together
    # (because shlex does not yet work with unicode...)
    return [item.replace('"', '') for item in 
        re.findall(r'[^\s"]*"[^"]*"[^\s"]*|[^\s"]+', cmd)]


def command(type):
    """Returns the command for the specified type as a list.
    
    Returns None if no command was specified.
    
    """
    cmd = QSettings().value("helper_applications/" + type, "")
    if not cmd:
        return
    
    if os.path.isabs(cmd) and os.access(cmd, os.X_OK):
        return [cmd]

    return shell_split(cmd)


def openUrl(url, type="browser"):
    """Opens the specified QUrl, using the specified type."""
    # be sure to pick a suitable type
    if url.scheme() == "mailto":
        type = "email"
    elif type == "browser" and url.toLocalFile():
        ext = os.path.splitext(url.toLocalFile())[1].lower()
        if os.path.isdir(url.toLocalFile()):
            type = "directory"
        elif ext == '.pdf':
            type = "pdf"
        elif ext in ('.png', '.jpg', '.jpeg'):
            type = "image"
        elif ext in ('.midi', '.mid'):
            type = "midi"
    
    # get the command
    cmd = command(type)
    if not cmd:
        if type != "shell":
            QDesktopServices.openUrl(url)
            return
        for cmd in terminalCommands():
            break # get the first
    
    prog = cmd.pop(0)
    
    workdir = None
    if url.toLocalFile():
        workdir = url.toLocalFile()
        if type != "shell":
            workdir = os.path.dirname(workdir)
    
    if any('$f' in a or '$u' in a for a in cmd):
        cmd = [a.replace('$u', url.toString())
                if '$u' in a else a.replace('$f', url.toLocalFile())
                for a in cmd]
    elif type in ("browser", "email"):
        cmd.append(url.toString())
    elif type != "shell":
        cmd.append(url.toLocalFile())
    
    try:
        subprocess.Popen([prog] + cmd, cwd=workdir)
    except OSError:
        QMessageBox.critical(None, _("Error"), _(
            "Could not start {program}.\n"
            "Please check path and permissions.").format(program=prog))


def terminalCommands():
    """Yields suitable commands to open a terminal/shell window.
    
    There is always yielded at least one, which is suitable as default.
    
    """
    if os.name == "nt":
        yield ['cmd.exe']
    elif sys.platform == 'darwin':
        yield ['open', '-a', 'Terminal', '$f']
    else:
        # find a default linux terminal
        paths = os.environ.get('PATH', os.defpath).split(os.pathsep)
        for cmd in (
            ['lxterminal', '--working-directory=$f'],
            ['xfce4-terminal', '--working-directory=$f'],
            ['konsole', '--workdir', '$f'],
            ['gnome-terminal', '--working-directory=$f'],
            ):
            for p in paths:
                if p:
                    prog = os.path.join(p, cmd[0])
                    if os.access(prog, os.X_OK):
                        yield cmd
        yield ['xterm']





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Manages widgets in the borders of a QScrollArea.
"""

from PyQt4.QtCore import QEvent, QObject, QPoint, QRect, QSize


LEFT, TOP, RIGHT, BOTTOM = 0, 1, 2, 3

class BorderLayout(QObject):
    
    # order in which the widget spaces are filled.
    # if top and bottom are first, they get the full width.
    order = TOP, BOTTOM, LEFT, RIGHT
    
    def __init__(self, scrollarea):
        super(BorderLayout, self).__init__(scrollarea)
        self._resizing = False
        self._margins = 0, 0, 0, 0
        self._widgets = ([], [], [], [])
        scrollarea.viewport().installEventFilter(self)
    
    @classmethod
    def get(cls, scrollarea):
        """Gets the BorderLayout for the given scrollarea.
        
        If None exists, creates and returns a new instance.
        
        """
        for c in scrollarea.children():
            if type(c) is cls:
                return c
        return cls(scrollarea)
    
    def scrollarea(self):
        """Returns our scrollarea instance (our parent)."""
        return self.parent()
    
    def addWidget(self, widget, side):
        """Adds a widget to our scrollarea."""
        self.insertWidget(widget, side, -1)
        
    def insertWidget(self, widget, side, position):
        """Inserts a widget in a specific position."""
        assert side in (LEFT, TOP, RIGHT, BOTTOM)
        new = False
        for l in self._widgets:
            if widget in l:
                l.remove(widget)
                break
        else:
            new = True
            widget.setParent(self.scrollarea())
            widget.installEventFilter(self)
        if position == -1:
            self._widgets[side].append(widget)
        else:
            self._widgets[side].insert(widget, position)
        if new:
            widget.show()
        self.updateGeometry()
            
    def removeWidget(self, widget):
        """Removes the widget, it is not deleted."""
        for l in self._widgets:
            if widget in l:
                l.remove(widget)
                widget.removeEventFilter(self)
                widget.setParent(None)
                self.updateGeometry()
    
    def setViewportMargins(self, left, top, right, bottom):
        """(Internal) Sets the viewport margins and remembers them."""
        self._margins = (left, top, right, bottom)
        try:
            self.scrollarea().setViewportMargins(left, top, right, bottom)
        except RuntimeError:
            # this can happen when the scrollarea already is deleted by Python
            # because setViewportMargins is a protected method
            pass
    
    def viewportGeometry(self):
        """(Internal) Returns the viewport geometry as if with 0 margins."""
        g = self.scrollarea().viewport().geometry()
        left, top, right, bottom = self._margins
        return g.adjusted(-left, -top, right, bottom)
        
    def eventFilter(self, obj, ev):
        """Reimplemented to handle resizes and avoid resize loops."""
        if self._resizing:
            return False
        elif ev.type() == QEvent.Resize and obj is self.scrollarea().viewport():
            self.updateGeometry()
        elif ev.type() in (QEvent.Resize, QEvent.ShowToParent, QEvent.HideToParent):
            self.updateGeometry()
        elif ev.type() in (QEvent.ParentChange, QEvent.DeferredDelete) and obj is not self.scrollarea().viewport():
            for l in self._widgets:
                if obj in l:
                    l.remove(obj)
                    obj.removeEventFilter(self)
                    self.updateGeometry()
                    break
        return False
    
    def updateGeometry(self):
        """(Internal) Positions all widgets in the scrollarea edges."""
        self._resizing = True
        g = self.viewportGeometry()
        pos = g.topLeft()
        size = g.size()
        left, right, top, bottom = 0, 0, 0, 0
        
        for side in self.order:
            def widgets():
                for w in self._widgets[side][::-1]:
                    if w.isVisibleTo(self.scrollarea()):
                        yield w
            if side is LEFT:
                for widget in widgets():
                    w = widget.sizeHint().width()
                    h = size.height() - top - bottom
                    x = left
                    y = top
                    g = QRect(pos + QPoint(x, y), QSize(w, h))
                    widget.setGeometry(g)
                    left += w
            elif side is RIGHT:
                for widget in widgets():
                    w = widget.sizeHint().width()
                    h = size.height() - top - bottom
                    x = size.width() - right - w
                    y = top
                    g = QRect(pos + QPoint(x, y), QSize(w, h))
                    widget.setGeometry(g)
                    right += w
            elif side is TOP:
                for widget in widgets():
                    w = size.width() - left - right
                    h = widget.heightForWidth(w)
                    if h == -1:
                        h = widget.sizeHint().height()
                    x = left
                    y = top
                    g = QRect(pos + QPoint(x, y), QSize(w, h))
                    widget.setGeometry(g)
                    top += h
            elif side is BOTTOM:
                for widget in widgets():
                    w = size.width() - left - right
                    h = widget.heightForWidth(w)
                    if h == -1:
                        h = widget.sizeHint().height()
                    x = left
                    y = size.height() - bottom - h
                    g = QRect(pos + QPoint(x, y), QSize(w, h))
                    widget.setGeometry(g)
                    bottom += h
        self.setViewportMargins(left, top, right, bottom)
        self._resizing = False






# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Named session support (not to be confused with the QSessionManager support
in session.py)

A session is a global list of open documents, with some additional preferences set.

"""

from __future__ import unicode_literals

import itertools

from PyQt4.QtCore import QSettings, QUrl

import app
import util
import qsettings


_currentSession = None


@app.mainwindowClosed.connect
def _saveLastUsedSession():
    s = QSettings()
    s.beginGroup("session")
    s.setValue("lastused", _currentSession or "")
    
def loadDefaultSession():
    """Load the session which should be started by default.
    
    This can be:
    - no session,
    - last used session,
    - a specific session.
    
    Returns the document that should be set active (if any).
    """
    s = QSettings()
    s.beginGroup("session")
    start = s.value("startup", "none", type(""))
    name = None
    if start == "lastused":
        name = s.value("lastused", "", type(""))
    elif start == "custom":
        name = s.value("custom", "", type(""))
        if name not in sessionNames():
            s.setValue("startup", "none")
    if name and name in sessionNames():
        return loadSession(name)

def sessionGroup(name):
    """Returns the session settings group where settings can be stored for the named session.
    
    If the group doesn't exist, it is created.
    
    """
    session = app.settings("sessions")
    childGroups = session.childGroups()
    for group in childGroups:
        if session.value(group + "/name", "", type("")) == name:
            break
    else:
        for count in itertools.count(1):
            group = "session{0}".format(count)
            if group not in childGroups:
                session.setValue(group +  "/name", name)
                break
    session.beginGroup(group)
    return session

def sessionNames():
    session = app.settings("sessions")
    names = [session.value(group + "/name", "", type("")) for group in session.childGroups()]
    names.sort(key=util.naturalsort)
    return names
    
def loadSession(name):
    """Loads the given session (without closing other docs first).
    
    Return the document that should become the active one.
    If None is returned, the session did not open any documents!
    
    """
    session = sessionGroup(name)
    urls = qsettings.get_url_list(session, "urls")
    active = session.value("active", -1, int)
    result = None
    docs = []
    for url in urls:
        try:
            doc = app.openUrl(url)
        except IOError:
            pass
        else:
            docs.append(doc)
    setCurrentSession(name)
    if docs:
        if active not in range(len(docs)):
            active = 0
        return docs[active]

def saveSession(name, documents, activeDocument=None):
    """Saves the list of documents and which one is active."""
    # only save the documents that have an url
    documents = [doc for doc in documents if not doc.url().isEmpty()]
    session = sessionGroup(name)
    session.setValue("urls", [doc.url() for doc in documents])
    if activeDocument in documents:
        session.setValue("active", documents.index(activeDocument))
    else:
        session.remove("active")
    app.saveSessionData(name)

def deleteSession(name):
    session = app.settings("sessions")
    for group in session.childGroups():
        if session.value(group + "/name", "", type("")) == name:
            session.remove(group)
            break
    if name == _currentSession:
        setCurrentSession(None)

def renameSession(old, new):
    """Renames a session."""
    session = sessionGroup(old)
    session.setValue("name", new)
    if old == currentSession():
        setCurrentSession(new)
    
def currentSession():
    return _currentSession
    
def setCurrentSession(name):
    global _currentSession
    if name != _currentSession:
        name and sessionGroup(name) # just select it, so its name is written in case it doesn't exist
        _currentSession = name
        app.sessionChanged(name)

def currentSessionGroup():
    """Returns the session settings at the current group is there is a current session.
    
    If there is no current session, returns None.
    
    """
    if _currentSession:
        return sessionGroup(_currentSession)





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Generate meaningful tooltips for cursor positions.
"""

from __future__ import unicode_literals

import tokeniter
import ly.lex.lilypond


def text(cursor):
    filename = cursor.document().documentName()
    line = cursor.blockNumber() + 1
    column = cursor.position() - cursor.block().position()
    text = "{0} ({1}:{2})".format(filename, line, column)
    definition = get_definition(cursor)
    if definition:
        text += '\n' + definition
    time_pos = time_position(cursor)
    if time_pos:
        text += '\n' + _("Position: {pos}").format(pos=time_pos)
    return text

def get_definition(cursor):
    block = cursor.block()
    while block.isValid():
        state = tokeniter.state(block)
        if isinstance(state.parser(), ly.lex.lilypond.ParseGlobal):
            for t in tokeniter.tokens(block)[:2]:
                if type(t) is ly.lex.lilypond.Name:
                    return t[:]
                elif isinstance(t, ly.lex.lilypond.Keyword) and t == '\\score':
                    return '\\score'
        block = block.previous()

def time_position(cursor):
    import documentinfo
    pos = documentinfo.music(cursor.document()).time_position(cursor.position())
    if pos is not None:
        import ly.duration
        return ly.duration.format_fraction(pos)





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Backup files before overwriting
"""

from __future__ import unicode_literals

import os
import shutil

from PyQt4.QtCore import QSettings


def backup(filename):
    """Makes a backup of 'filename'.
    
    Returns True if the backup succeeded.
    
    """
    if filename:
        try:
            shutil.copy(filename, backupName(filename))
            return True
        except (IOError, OSError):
            pass
    return False


def removeBackup(filename):
    """Removes filename's backup unless the user has configured to keep it."""
    if filename and not QSettings().value("backup_keep", False, bool):
        try:
            os.remove(backupName(filename))
        except (IOError, OSError):
            pass


def scheme():
    """Returns a string that must contain "FILE".
    
    Replacing that part yields the backup name.
    
    """
    s = QSettings().value("backup_scheme", "FILE~")
    assert 'FILE' in s and s != 'FILE'
    return s


def backupName(filename):
    """Returns the backup file name for the given filename."""
    return scheme().replace("FILE", filename)





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Some useful tools dealing with popplerqt4 (PDF) documents.
"""

import os


class Document(object):
    """Represents a (lazily) loaded PDF document."""
    def __init__(self, filename=''):
        self._filename = filename
        self._document = None
        self._dirty = True
        
    def filename(self):
        """Returns the filename, set on init or via setFilename()."""
        return self._filename
    
    def setFilename(self, filename):
        """Sets a filename.
        
        The document will be reloaded next time it is requested.
        
        """
        self._filename = filename
        self._dirty = True
            
    def name(self):
        """Returns the filename without path."""
        return os.path.basename(self._filename)
        
    def document(self):
        """Returns the PDF document the filename points to, reloading if the filename was set.
        
        Can return None, in case the document failed to load.
        
        """
        if self._dirty:
            self._document = self.load()
            self._dirty = False
        return self._document

    def load(self):
        """Should load and return the popplerqt4 Document for our filename."""
        try:
            import popplerqt4
        except ImportError:
            return
        return popplerqt4.Poppler.Document.load(self._filename)





import os
import sys
from frescobaldi_app import appinfo

try:
    from setuptools import setup
    USE_SETUPTOOLS = True
except ImportError:
    from distutils.core import setup
    USE_SETUPTOOLS = False


def packagelist(directory):
    """Returns a sorted list with package names for all packages under the given directory."""
    return list(sorted(root.replace(os.sep, '.')
        for root, dirs, files in os.walk(directory)
        if '__init__.py' in files))

scripts = ['frescobaldi']
packages = packagelist('frescobaldi_app')
package_data = {
    'frescobaldi_app.css': ['*.png'],
    'frescobaldi_app.help': ['*.png'],
    'frescobaldi_app.hyphdicts': ['*.dic'],
    'frescobaldi_app.icons': [
        '*.ico',
        '*.svg',
        '*x*/*.png',
        'Tango/index.theme',
        'Tango/scalable/*.svg',
        'TangoExt/index.theme',
        'TangoExt/scalable/*.svg',
    ],
    'frescobaldi_app.layoutcontrol': ['*.ly', '*.ily'],
    'frescobaldi_app.po': ['*.mo'],
    'frescobaldi_app.scorewiz': ['*.png'],
    'frescobaldi_app.splashscreen': ['*.png'],
    'frescobaldi_app.svgview': ['*.js', '*.html'],
    'frescobaldi_app.symbols': ['*.svg'],
    'frescobaldi_app.userguide': ['*.md', '*.png'],
}
options = {
    'sdist': {
        'force_manifest': 1,
    }
}

if sys.platform.startswith('win'):
    scripts.append('windows/frescobaldi-wininst.py')
    options['bdist_wininst'] = {
        'install_script': 'windows/frescobaldi-wininst.py',
        'bitmap': 'windows/frescobaldi-wininst.bmp',
    }
    data_files = []
else:
    data_files = [
        ('share/icons/hicolor/scalable/apps', ['frescobaldi_app/icons/frescobaldi.svg']),
        ('share/applications', ['frescobaldi.desktop']),
        ('share/man/man1', ['frescobaldi.1']),
    ]

setup_extra_args = {}
if USE_SETUPTOOLS:
    setup_extra_args['install_requires'] = ['python-ly', 'python-poppler-qt4']

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Environment :: MacOS X',
    'Environment :: Win32 (MS Windows)',
    'Environment :: X11 Applications :: Qt',
    'Intended Audience :: End Users/Desktop',
    'License :: OSI Approved :: GNU General Public License (GPL)',
    # Natural Language :: Chinese (Hong Kong) is not yet accepted by pypi
    #'Natural Language :: Chinese (Hong Kong)',
    'Natural Language :: Chinese (Simplified)',
    'Natural Language :: Chinese (Traditional)',
    'Natural Language :: Czech',
    'Natural Language :: Dutch',
    'Natural Language :: English',
    'Natural Language :: French',
    'Natural Language :: Galician',
    'Natural Language :: German',
    'Natural Language :: Italian',
    'Natural Language :: Polish',
    'Natural Language :: Portuguese (Brazilian)',
    'Natural Language :: Russian',
    'Natural Language :: Spanish',
    'Natural Language :: Turkish',
    'Natural Language :: Ukranian',
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.2',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Topic :: Multimedia :: Sound/Audio',
    'Topic :: Multimedia :: Graphics',
    'Topic :: Text Editors',
]

setup(
    name = appinfo.name,
    version = appinfo.version,
    description = appinfo.description,
    long_description = appinfo.long_description,
    maintainer = appinfo.maintainer,
    maintainer_email = appinfo.maintainer_email,
    url = appinfo.url,
    license = appinfo.license,
    
    scripts = scripts,
    packages = packages,
    package_data = package_data,
    data_files = data_files,
    classifiers = classifiers,
    options = options,
    **setup_extra_args
)




# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
The global menubar on Mac OS X.

This menubar is only created when there is no MainWindow. As soon a MainWindow
is created, this global menubar is deleted, and Mac OS X will use the menubar
from the MainWindow.

"""

from __future__ import unicode_literals

import sys
import os

from PyQt4.QtCore import QUrl
from PyQt4.QtGui import QAction, QFileDialog, QMenu, QMenuBar

import app
import util
import qutil
import icons
import appinfo

from . import use_osx_menu_roles


def setup():
    """Create the global menu bar, and switch to no-session."""
    global _menubar
    _menubar = menubar()
    import sessions
    sessions.setCurrentSession(None)

@app.mainwindowCreated.connect
def delete():
    """Delete the global menu bar."""
    global _menubar
    _menubar.deleteLater()
    _menubar = None

def menubar():
    """Return a newly created parent-less menu bar that's used when there is no main window."""
    m = QMenuBar()
    m.addMenu(menu_file(m))
    m.addMenu(menu_edit(m))
    m.addMenu(menu_window(m))
    m.addMenu(menu_sessions(m))
    m.addMenu(menu_help(m))
    return m

def menu_file(parent):
    m = QMenu(parent)
    m.setTitle(_("menu title", "&File"))
    m.addAction(icons.get('document-new'), _("action: new document", "&New"), file_new)
    m.addMenu(menu_file_new_from_template(m))
    m.addAction(icons.get('tools-score-wizard'), _("New Score with &Wizard..."), file_new_with_wizard)
    m.addSeparator()
    m.addAction(icons.get('document-open'), _("&Open..."), file_open)
    m.addMenu(menu_file_open_recent(m))
    m.addSeparator()
    m.addMenu(menu_file_import(m))
    m.addSeparator()
    role = QAction.QuitRole if use_osx_menu_roles() else QAction.NoRole
    m.addAction(icons.get('application-exit'), _("&Quit"), app.qApp.quit).setMenuRole(role)
    return m

def menu_file_new_from_template(parent):
    m = QMenu(parent)
    m.setTitle(_("New from &Template"))
    m.triggered.connect(slot_file_new_from_template_action)
    from snippet import model, actions, snippets
    groups = {}
    for name in sorted(model.model().names()):
        variables = snippets.get(name).variables
        group = variables.get('template')
        if group:
            action = actions.action(name, m)
            if action:
                groups.setdefault(group, []).append(action)
    for group in sorted(groups):
        for action in groups[group]:
            m.addAction(action)
        m.addSeparator()
    qutil.addAccelerators(m.actions())
    return m

def menu_file_open_recent(parent):
    m = QMenu(parent)
    m.setTitle(_("Open &Recent"))
    m.triggered.connect(slot_file_open_recent_action)
    import recentfiles
    for url in recentfiles.urls():
        f = url.toLocalFile()
        dirname, basename = os.path.split(f)
        text = "{0}  ({1})".format(basename, util.homify(dirname))
        m.addAction(text).url = url
    qutil.addAccelerators(m.actions())
    return m

def menu_file_import(parent):
    m = QMenu(parent)
    m.setTitle(_("submenu title", "&Import"))
    m.addAction(_("Import MusicXML..."), file_import_musicxml)
    return m
    
def menu_edit(parent):
    m = QMenu(parent)
    m.setTitle(_("menu title", "&Edit"))
    role = QAction.PreferencesRole if use_osx_menu_roles() else QAction.NoRole
    m.addAction(icons.get('preferences-system'), _("Pr&eferences..."), edit_preferences).setMenuRole(role)
    return m

def menu_window(parent):
    m = QMenu(parent)
    m.setTitle(_('menu title', '&Window'))
    m.addAction(icons.get('window-new'), _("New &Window"), file_new)
    return m

def menu_sessions(parent):
    m = QMenu(parent)
    m.setTitle(_('menu title', '&Session'))
    m.triggered.connect(slot_session_action)
    import sessions
    for name in sessions.sessionNames():
        a = m.addAction(name.replace('&', '&&'))
        a.setObjectName(name)
    qutil.addAccelerators(m.actions())
    return m    
    
def menu_help(parent):
    m = QMenu(parent)
    m.setTitle(_('menu title', '&Help'))
    role = QAction.AboutRole if use_osx_menu_roles() else QAction.NoRole
    m.addAction(icons.get('help-about'), _("&About {appname}...").format(appname=appinfo.appname), help_about).setMenuRole(role)
    return m

def mainwindow():
    """Create, show() and return a new MainWindow."""
    import mainwindow
    w = mainwindow.MainWindow()
    w.show()
    w.activateWindow()
    return w

def file_new():
    mainwindow().newDocument()

def file_new_with_wizard():
    w = mainwindow()
    w.newDocument()
    import scorewiz
    scorewiz.ScoreWizard.instance(w).showDialog()

def file_open():
    filetypes = app.filetypes('.ly')
    caption = app.caption(_("dialog title", "Open File"))
    directory = app.basedir()
    files = QFileDialog.getOpenFileNames(None, caption, directory, filetypes)
    if files:
        w = mainwindow()
        w.openUrls([QUrl.fromLocalFile(f) for f in files])
        if not app.documents:
            w.cleanStart()

def slot_file_new_from_template_action(action):
    name = action.objectName()
    d = app.openUrl(QUrl())
    win = mainwindow()
    win.setCurrentDocument(d)
    from snippet import insert, snippets
    view = win.currentView()
    view.setFocus()
    insert.insert(name, view)
    d.setUndoRedoEnabled(False)
    d.setUndoRedoEnabled(True) # d.clearUndoRedoStacks() only in Qt >= 4.7
    d.setModified(False)
    if 'template-run' in snippets.get(name).variables:
        import engrave
        engrave.engraver(win).engrave('preview', d)

def slot_file_open_recent_action(action):
    w = mainwindow()
    d = w.openUrl(action.url)
    if d:
        w.setCurrentDocument(d)

def file_import_musicxml():
    w = mainwindow()
    w.newDocument()
    import file_import
    file_import.FileImport.instance(w).importMusicXML()

def edit_preferences():
    import preferences
    # TODO: make it possible to run Preferences without a Main Window.
    # Currently the Keyboard Shortcuts section needs the mainwindow to get
    # the current shortcuts.
    w = mainwindow()
    w.newDocument()
    preferences.PreferencesDialog(w).exec_()

def slot_session_action(action):
    name = action.objectName()
    import sessions
    doc = sessions.loadSession(name) or app.openUrl(QUrl())
    w = mainwindow()
    w.setCurrentDocument(doc)

def help_about():
    import about
    about.AboutDialog(None).exec_()





# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2008 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Repository of Frescobaldi application.
"""

from __future__ import unicode_literals

import sys
import os

from .gitrepo import GitRepo

class AppRepo(GitRepo):
    """
    Subclass to be used for the Frescobaldi Git repository.
    Offers methods exclusively useful for this specific repo.
    """
    def __init__(self):
        super(AppRepo, self).__init__((os.path.normpath(
                                        os.path.join(sys.path[0], '..'))))
        self._activeBranch = self.current_branch()
    
    def active_branch(self):
        """
        Returns the name of the branch that was current_branch
        when the application has started.
        current_branch() may of course change during the
        runtime of the application
        """
        return self._activeBranch



# This file is part of the Frescobaldi project, http://www.frescobaldi.org/
#
# Copyright (c) 2011 - 2014 by Wilbert Berendsen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
# See http://www.gnu.org/licenses/ for more information.

"""
Implementation of the tools to edit durations of selected music.

Durations are represented simply by lists of ly.lex.lilypond.Duration tokens.
"""

from __future__ import unicode_literals

import itertools

import icons
import inputdialog
import lydocument
import ly.rhythm


_clipboard = [] # clipboard for rhythm copy and paste

_history = set() # earlier rhythms typed in apply dialog


def rhythm_double(cursor):
    ly.rhythm.rhythm_double(lydocument.cursor(cursor))

def rhythm_halve(cursor):
    ly.rhythm.rhythm_halve(lydocument.cursor(cursor))

def rhythm_dot(cursor):
    ly.rhythm.rhythm_dot(lydocument.cursor(cursor))

def rhythm_undot(cursor):
    ly.rhythm.rhythm_undot(lydocument.cursor(cursor))

def rhythm_remove_scaling(cursor):
    ly.rhythm.rhythm_remove_scaling(lydocument.cursor(cursor))

def rhythm_remove_fraction_scaling(cursor):
    ly.rhythm.rhythm_remove_fraction_scaling(lydocument.cursor(cursor))

def rhythm_remove(cursor):
    ly.rhythm.rhythm_remove(lydocument.cursor(cursor))

def rhythm_implicit(cursor):
    ly.rhythm.rhythm_implicit(lydocument.cursor(cursor))

def rhythm_implicit_per_line(cursor):
    ly.rhythm.rhythm_implicit_per_line(lydocument.cursor(cursor))

def rhythm_explicit(cursor):
    ly.rhythm.rhythm_explicit(lydocument.cursor(cursor))

def rhythm_apply(cursor, mainwindow):
    durs = inputdialog.getText(mainwindow,
        _("Apply Rhythm"), _("Enter a rhythm:"),
        complete = sorted(_history),
        regexp = r'([0-9./* ]|\\breve|\\longa|\\maxima)+',
        help = "rhythm", icon = icons.get('tools-rhythm'))
    if not durs:
        return # user cancelled dialog
    durations = durs.split()
    if durations:
        _history.add(durs.strip())
        ly.rhythm.rhythm_overwrite(lydocument.cursor(cursor), durations)

def rhythm_copy(cursor):
    _clipboard[:] = ly.rhythm.rhythm_extract(lydocument.cursor(cursor))

def rhythm_paste(cursor):
    ly.rhythm.rhythm_overwrite(lydocument.cursor(cursor), _clipboard)

